(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Abstract Syntactic Tree *)

(* Capacities *)
type cap = In | In_ | Out | Out_ | Open | Open_ | Print | Pause of int ref

(* Recursion variables *)
type var = string

(* Expressions *)
type expr =
  | Name of string 
  | Capa of cap * string
  | Sleep of int

(* Paths *)
type path = expr list

(* Processus *)
type process = process_t list
  
(* Single processus *)
and process_t =
  | Pref of expr * process
  | Amb  of string * process
  | Amb' of (string * int) * string * process
  | Send of path list
  | Nu   of string * process
  | Rcv  of string list * process 
  | Rec  of var * process 
  | Var  of var

(* curry_nus p l = (nu l0)(nu l1)...(nu ln)p *)
let curry_nus p = function
  | [] -> p
  | w::q -> [List.fold_right (fun n p -> Nu(n, [p])) q (Nu(w, p))]

(* uncurry_nus (nu l0)(nu l1)...(nu ln)p = l,p*)
let uncurry_nus = 
  let rec aux acc = function
    | Nu(x,[y]) -> aux (x::acc) y
    | Nu(x,q) -> List.rev (x::acc), q
    | p -> List.rev acc, [p]
  in aux []

(* prefix l p = l0.l1...ln.p *)
let rec prefix l p = match l with
  | [] -> p
  | c::q -> [Pref(c, prefix q p)]

(* par p q = p|q *)
let par = List.rev_append

(* zero = 0 *)
let zero = []


(* recursion variables *)
let var x = x
let bang_var = "!"
let string_of_var x = x

(* abstraction variables *)
let avar x = "~"^x
let is_abstract x = x<>"" && x.[0]='~'

(* root name *)
let rootname = "R@@t"

(* is an ambient immobile or not : TMP *)
let immobile n = match n.[0] with 
  | 'A'..'Z' -> true
  | _ -> false


(* capacities *)
let string_of_cap = function
  | In    -> "in "
  | In_   -> "in_"
  | Out   -> "out "
  | Out_  -> "out_"
  | Open  -> "open "
  | Open_ -> "open_"
  | Print -> "print "
  | Pause r when !r=0 -> "pause "
  | Pause r           -> "pause "^string_of_int !r^" "


(* Tokens (for lexing) *)
type token =
  | COMMA
  | SEMI
  | LET
  | EQUAL
  | QUOTE
  | COLON
  | DOT
  | LPAR
  | RPAR
  | LSQR
  | RSQR
  | LPIK
  | RPIK
  | STAR
  | BAR
  | NU
  | REC
  | SLEEP
  | BANG
  | ZERO
  | EOF
  | CAPA  of cap
  | WORD  of string
  | HOST  of string
  | DIESE of string
  | INT   of int

(* parse errors *)
type error = int*int*string
let string_of_error (b,e,m) =
  Printf.sprintf "line %d, character %d : %s" b e m
exception ParseError of error


(* misc utilities *)

(* option with default *)
let oget o x = match o with
  | Some o -> o
  | None -> x

(* list printing *)
let string_of_list ?(sep=",") f = 
  let rec aux = function
    | [] -> ""
    | [x] -> f x
    | x::q -> f x ^ sep ^ aux q
  in aux


(* formatter converter *)
let (!!) f = (fun () x -> f Format.str_formatter x; Format.flush_str_formatter())

(* user-friendly failwith *)
let failwith x = Format.kprintf failwith x
