(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Abstract Syntactic Tree *)

(* Capacities *)
type cap = In | In_ | Out | Out_ | Open | Open_ | Print | Pause of int ref 

(* Recursion variables *)
type var

(* Expressions *)
type expr = 
  | Name of string
  | Capa of cap * string
  | Sleep of int

(* Paths *)
type path = expr list

(* Processus *)
type process = process_t list

(* Single processus *)
and process_t =
  | Pref of expr * process
  | Amb  of string * process
  | Amb' of (string * int) * string * process
  | Send of path list
  | Nu   of string * process
  | Rcv  of string list * process
  | Rec  of var * process
  | Var  of var


(* curry_nus p l = (nu l0)(nu l1) ... (nu ln)p *)
val curry_nus  : process -> string list -> process

(* uncurry_nus (nu l0)(nu l1) ... (nu ln)p = l,p*)
val uncurry_nus: process_t -> string list * process

(* prefix l p = l0.l1. ... ln.p *)
val prefix: path -> process -> process

(* par p q = p|q *)
val par: process -> process -> process

(* zero = 0 *)
val zero: process


(* name of the ambient toplevel *)
val rootname: string


(* temporary function to determine the type of an ambient name *)
val immobile: string -> bool


(* recursion variables *)
val var: string -> var
val bang_var: var
val string_of_var: var -> string

(* abstraction variables *)
val avar: string -> string
val is_abstract: string -> bool

(* direct conversion *)
val string_of_cap : cap -> string



(* Tokens (for lexing) *)
type token =
  | COMMA
  | SEMI
  | LET
  | EQUAL
  | QUOTE
  | COLON
  | DOT
  | LPAR
  | RPAR
  | LSQR
  | RSQR
  | LPIK
  | RPIK
  | STAR
  | BAR
  | NU
  | REC
  | SLEEP
  | BANG
  | ZERO
  | EOF
  | CAPA  of cap
  | WORD  of string
  | HOST  of string
  | DIESE of string
  | INT   of int

(* Parse errors *)
type error = int*int*string
val string_of_error: error -> string
exception ParseError of error



(* misc utilities *)

(* option with default *)
val oget: 'a option -> 'a -> 'a 

(* list printing *)
val string_of_list: ?sep:string -> ('a -> string) -> 'a list -> string

(* formatter converter *)
val ( !! ) : (Format.formatter -> 'a -> 'b) -> unit -> 'a -> string

(* user-friendly failwith *)
val failwith : ('a, unit, string, 'b) format4 -> 'a 
