(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Communications of the GC-PAN *) 

open Ast

(* pre-matches *)
type match_ =
    [ `Com   of string list * path list * process
    | `Print of string * process
    | `Pause of int ref * string * process
    | `Sleep of int * process ] 


(* requests *)
type 'c request_ = 
    [ `In    of string * 'c
    | `In_   of string * 'c
    | `Open_ of string * 'c
    | `Out   of string * 'c
    | `DoOut of 'c * 'c
    ]

(* packed requests *)
and 'c prequest_ = 'c request_ list * int * 'c list
      
(* matches (for states) *)
and 'c matsh =
    [ match_
    | `None
    | `Req   of 'c request_ * process
    | `In    of 'c * 'c * string
    | `Out   of 'c * process 
    | `DoOut of 'c * 'c 
    | `Open  of process * 'c ]
and 'c realmatch =
    [ `Com   of string list * path list * process
    | `Print of string * process 
    | `Pause of int ref * string * process 
    | `Sleep of int * process 
    | `Req   of 'c request_ * process
    | `In    of 'c * 'c * string
    | `Out   of 'c * process 
    | `Open  of process * 'c]

(* frozen guards/requests (their names are not the current name) 
 * (for states) *)
type frozen_ = 
    [ `Out_    of string * process 
    | `R_in_   of string * process
    | `R_open_ of string * process
    ]
type 'c frozen =
    [ frozen_
    | `Out of string * 'c]

(* matchers for Local rules (for states) *)
module ID = struct
  type 'c t = 'c
  let conv f = f
  let format g = g
end
module ID2 = struct
  type 'c t = 'c * string
  let conv f (c,s) = f c, s
  let format g f (c,s) = Format.fprintf f "%a,%s" g c s
end
module CONST(M: sig type t val format: Format.formatter -> t -> unit end) = struct
  type 'c t = M.t
  let conv f x = x
  let format g = M.format
end
module PATH=CONST(struct type t = path list           let format = Printer.format_list Printer.format_path end)
module PROC=CONST(struct type t = process             let format = Printer.format_process end)
module ABST=CONST(struct type t = string list*process let format f (x,p) = Printer.format_process_t f (Rcv(x,p)) end)
module Coms  = Matcher.Simple(PATH)(ABST)
  (struct 
     type 'c t = 'c realmatch 
     let mix p (x,q) = `Com (x,p,q) 
   end)
module Outs  = Matcher.Simple(ID)(PROC)
  (struct 
     type 'c t = 'c realmatch 
     let mix h p = `Out (h,p) 
   end)
module Ins   = Matcher.SimpleNamed(ID)(ID2)
  (struct 
     type 'c t = 'c realmatch 
     let mix h (k,s) = `In (h,k,s) 
   end)
module Opens = Matcher.SimpleNamed(PROC)(ID)
  (struct 
     type 'c t = 'c realmatch 
     let mix p h = `Open (p,h)
   end)

(* state of a located ambient *) 
type 'c state_ = {
  name:  string;
  coms:  'c Coms.t;
  outs:  'c Outs.t;
  ins:   'c Ins.t;
  opens: 'c Opens.t;
  doouts: ('c * 'c) list;
  frozen:  'c frozen list;
  matches: 'c realmatch list;
}

(* collectors *)
class type ['c] collector_ =
object
  method none:            unit
  method forwarder:       'c -> int -> 'c -> unit
  method bforwarder:      'c -> int -> 'c -> 'c option -> 'c prequest_ list -> unit
  method st_ambient:      'c -> int -> 'c -> 'c state_ -> unit
  method wait_st_ambient: 'c -> int -> 'c -> 'c state_ -> ?w:string -> process -> unit
  method im_ambient:      'c -> int -> 'c -> 'c state_ -> unit
end

(* completions *)
type 'c completion_ =
    [ `Go       of 'c
    | `Gof	of 'c
    | `OKin	of 'c
    | `Migrate  of 'c
    | `Register of bool * process * 'c state_ ]

(* maintenance messages *)
type local_command =
    [ `Eval of process 
    | `Step of string
    | `Collect of Tree.collector
    | `Debug
    | `Stop ]

(* messages *)
type 'c message_ =
    [ `R of 'c prequest_
    | `C of 'c completion_
    | `Decr 
    | local_command ]

(* commands *)
type command__ =
    [ `Eval of process 
    | `Clear
    | `Stop
    | `Step of string
    | `Debug
    | `Tree ]
type 'c command_ =
    [ command__
    | `Create of string * process * 'c 
    | `Register of 'c ]


(* build the communication primitives with GenComm *)
include GenComm.Make
  (struct
     type 'c t = 'c message_
	 
     let req_conv f = function
       | `In    (n,c)  -> `In    (n,f c)
       | `In_   (n,c)  -> `In_   (n,f c)
       | `Out   (n,c)  -> `Out   (n,f c)
       | `Open_ (n,c)  -> `Open_ (n,f c)
       | `DoOut (c,c') -> `DoOut (f c,f c')
     let match_conv f = function
       | `Req   (r,p)    -> `Req   (req_conv f r,p) 
       | `In    (c,c',n) -> `In    (f c,f c',n)   
       | `Out   (c,p)    -> `Out   (f c,p)    
       | `Open  (p,c)    -> `Open  (p,f c)    
       | #match_ as x    -> x
     let frozen_conv f = function
       | `Out (s,c)    -> `Out (s,f c)
       | #frozen_ as x -> x
     let state_conv f s = {
       name    = s.name;
       coms    = Coms.conv  f s.coms;
       outs    = Outs.conv  f s.outs;
       ins     = Ins.conv   f s.ins;
       opens   = Opens.conv f s.opens;
       doouts  = List.rev_map (fun (c,c') -> f c,f c') s.doouts;
       frozen  = List.rev_map (frozen_conv f) s.frozen;
       matches = List.rev_map (match_conv  f) s.matches;
     }	   
     let compl_conv f = function
       | `Go        c      -> `Go       (f c)
       | `Gof	    c      -> `Gof      (f c) 
       | `OKin	    c      -> `OKin     (f c) 
       | `Migrate   c      -> `Migrate  (f c)
       | `Register (b,p,s) -> `Register (b,p,state_conv f s) 
     let conv f = function
       | `R (b,j,s) -> `R (List.rev_map (req_conv f) b, j, List.rev_map f s)
       | `C c       -> `C (compl_conv f c)       
       | `Decr      -> `Decr
       | #local_command as x -> x
   end)
   (struct
      type 'c t = 'c command_
      let conv f = function
	| `Create (n,p,s) -> `Create (n,p,f s)
	| `Register c     -> `Register (f c)
	| #command__ as x -> x
    end)


(* now we define exported types with the given type of channels *)
type request    = channel request_
type prequest   = channel prequest_
type completion = channel completion_
type state      = channel state_




(* -- Communication functions -- *)

(* restricted send/broadcast *)
let send_compl c m = send c (`C m)
let send_eval  c p = send c (`Eval p)
let send_req   c r = send c (`R ([r], 0, []))
let send_preq  c r = send c (`R r)
let send_decr  c   = send c `Decr
let local_broadcast m = local_broadcast (m:>message)

(* restricted channels/master *)
let channel f = snd (channel f)
let master  f = ignore (master f)

(* purger for Stopped agents *)
let stop_purge = 
  `Close (function
	    | `Collect o -> o#dead
	    | `Debug     -> Tree.debug "stopped agent"
	    | #message   -> ())

(* broadcast of a Gof message *)
let broadcast_gof l h = 
  List.iter (fun e -> Stats.relocation(); send_compl e (`Gof h)) l


(* collector for the local net *)
let collect = Tree.collect local_broadcast



(* -- Names handling -- *)

(* fresh names depends on the uid of the local site *)
let safe_names = ref true
let fresh_name =
  let m = Mutex.create () in
  let t = Hashtbl.create 12 in
    fun n -> 
      Mutex.lock m;
      let r = 
	try Hashtbl.find t n 
	with Not_found -> 
	  let r = ref (-1) in
	    Hashtbl.add t n r;
	    r
      in
	incr r;
	Mutex.unlock m;
	if !safe_names then n^masterid^string_of_int !r
	else n^"#"^string_of_int !r




(* -- States handling -- *)

let add_freeze s o = { s with frozen  = o::s.frozen  }
let add_match  s m = { s with matches = m::s.matches }

let add_out  s (n,h as nh) = 
  if n=s.name then 
    let o,m = Outs.add_left  (s.outs,s.matches) h in
      { s with outs=o; matches=m }
  else add_freeze s (`Out nh)

let add_out_ s (n,p as np) = 
  if n=s.name then 
    let o,m = Outs.add_right (s.outs,s.matches) p in
      { s with outs=o; matches=m }
  else add_freeze s (`Out_ np)

let add_r_in_ s n p h =
  if n=s.name then add_match s (`Req (`In_ (n, h), p))
  else add_freeze s (`R_in_ (n, p))

let add_r_open_ s n p h =
  if n=s.name then add_match s (`Req (`Open_ (n, h), p))
  else add_freeze s (`R_open_ (n, p))

let add_pref s c n p h = match c with
    | In      -> add_match   s (`Req (`In  (n, h), p))
    | Out     -> add_match   s (`Req (`Out (n, h), p))
    | Print   -> add_match   s (`Print (n, p))
    | Pause r -> add_match   s (`Pause (r, n, p))
    | Out_    -> add_out_    s (n, p) 
    | In_     -> add_r_in_   s n p h
    | Open_   -> add_r_open_ s n p h
    | Open    -> let o,m = Opens.add_left  (s.opens,s.matches) (n,p) in { s with opens=o; matches=m }

let add_sleep s i p = add_match s (`Sleep (i, p))

(* empty state, with the given name *)
let empty_state n = {
  name    = n;
  coms    = Coms.empty;
  outs    = Outs.empty;
  ins     = Ins.empty;
  opens   = Opens.empty;
  doouts  = [];
  frozen  = [];
  matches = [];
}

(* add a request to a state *)
let add_req s = function
  | `Out   nh   -> add_out   s nh
  | `In_  (n,h) -> let i,m = Ins.add_right   (s.ins,s.matches)   (n,(h,n)) in { s with ins=i; matches=m }
  | `In    nh   -> let i,m = Ins.add_left    (s.ins,s.matches)   nh in { s with ins=i; matches=m }
  | `Open_ nh   -> let o,m = Opens.add_right (s.opens,s.matches) nh in { s with opens=o; matches=m }
  | `DoOut hk   -> { s with doouts=hk::s.doouts } 
let add_reqs = List.fold_left add_req
      
(* register another state *)
let merge s h t = 
  let m = s.matches in
  let coms, m = Coms.add  (s.coms, m) t.coms  in
  let ins,  m = Ins.add   (s.ins,  m) t.ins   in
  let opens,m = Opens.add (s.opens,m) t.opens in
  let (@) = List.rev_append in
    if s.name = t.name then
      let outs,m = Outs.add (s.outs, m) t.outs in
        { name    = s.name;
	  coms    = coms;
	  ins     = ins;
	  opens   = opens;
	  outs    = outs;
	  doouts  = s.doouts @ t.doouts;
	  frozen  = s.frozen @ t.frozen;
	  matches = t.matches @ m }
    else
      let f = 
	(* freezage de tous les t.outs *)
	List.fold_left (fun acc p -> `Out_ (t.name,p) :: acc)
	  (List.fold_left (fun acc h -> `Out (t.name,h) :: acc) 
	     s.frozen
	     (Outs.get_left t.outs))
	  (Outs.get_right t.outs)
      in
      let s =
	(* dfreezage des bons outs/req de t.frozen*)
	List.fold_left 
	  (fun s -> function
	     | `Out      nh   -> add_out     s nh
	     | `Out_     np   -> add_out_    s np
	     | `R_in_   (n,p) -> add_r_in_   s n p h
	     | `R_open_ (n,p) -> add_r_open_ s n p h)
	  { name    = s.name;
	    coms    = coms;
	    ins     = ins;
	    opens   = opens;
	    outs    = s.outs;
	    doouts  = s.doouts @ t.doouts;
	    frozen  = f;
	    matches = m } 
	  t.frozen
      in
      let f, m =
	(* filtrage de t.matches *)
	List.fold_left 
	  (fun (f,m) -> function
	     | `Out (h,p)            -> (`Out_ (t.name,p)::`Out (t.name,h)::f,m)
	     | `Req (`In_  (n,_), p) -> (`R_in_(n,p)::f,m)
	     | `Req (`Open_(n,_), p) -> (`R_open_(n,p)::f,m)
	     |  x                    -> (f,x::m)
	  )
	  (s.frozen, s.matches)
	  t.matches
      in
        { s with frozen=f; matches=m }
(*
t.outs    --------> frozen
t.frozen  -filter-> frozen / outs -> matches
t.matches -filter-> frozen / matches
*)

(* add a process to a state *)
let store s ~(create: ?host:GenComm.host -> ?h:channel -> ?n:string -> process -> channel*int) h i p = 
  let add_send s e = let c,m = Coms.add_left  (s.coms,s.matches) e in
    { s with coms=c; matches=m }
  in
  let add_rcv  s a = let c,m = Coms.add_right (s.coms,s.matches) a in
    { s with coms=c; matches=m }
  in
  let rec aux (i,s as is) = function
    | Pref(Capa(c,n),p) -> i, add_pref s c n p h
    | Pref(Sleep t,  p) -> i, add_sleep s t p
    | Send e            -> i, add_send s e
    | Rcv (x,p)         -> i, add_rcv s (x,p)
    | Amb (n,p)         -> i + snd (create ~h ~n p), s
    | Amb'(host,n,p)    -> send_command ~host (`Create (n,p,h)); i+1,s
    | Nu  (n,p)         -> auxl is (Subst.name n (fresh_name n) p)
    | Rec (x,p) as q    -> auxl is (Subst.process_t x q p)
    | Pref(Name v,_)    -> failwith "unbound path variable (%s)" v
    | Var x             -> failwith "unbound rec variable (%s)" (string_of_var x)
  and auxl is p = List.fold_left aux is p
  in auxl (i,s) p


(* get a match from the state *)
let get_match s = 
  match s.doouts with 
    | x::q -> { s with doouts=q }, `DoOut x
    | [] -> match s.matches with
	| [] -> s,`None
	| x::q -> { s with matches=q }, (x:> _ matsh)
	    
(* get an async match from the state (for waiting ST) *)
let get_async_match s = 
  let rec rem acc = function
    | [] -> None
    | `In x :: q -> Some (List.rev_append acc q, x)
    | x::q -> rem (x::acc) q
  in 
    match s.doouts with 
      | x::q -> { s with doouts=q }, `DoOut x
      | [] -> match rem [] s.matches with
	  | Some (m,x) -> {s with matches=m}, `In x
	  | None       ->  s, `None

(* get a non "`Req" match from the state (for Immobiles) *)
type nreqmatch =
  [ `Com   of string list * path list * process 
  | `In    of channel * channel * string
  | `Out   of channel * process 
  | `Open  of process * channel
  | `Print of string * process
  | `Pause of int ref * string * process
  | `Sleep of int * process ]
let get_nonReq_match s = 
  let rec rem acc = function
    | [] -> None
    | `Req _ as x :: q -> rem (x::acc) q
    | #nreqmatch as x :: q -> Some (List.rev_append acc q, x)
  in 
    match s.doouts with 
      | x::q -> { s with doouts=q }, `DoOut x
      | [] -> match rem [] s.matches with
	  | Some (m,x) -> {s with matches=m}, x
	  | None       -> s, `None

(* get the name of a state *)
let state_name s = s.name



(* -- Pretty-printing -- *)

let string_of_request = function
  | `In    (n,h) -> "In "  ^n^","^string_of_channel h
  | `Out   (n,h) -> "Out " ^n^","^string_of_channel h
  | `DoOut (h,k) -> "DoOut " ^string_of_channel h^","^string_of_channel k
  | `In_   (n,h) -> "In_"  ^n^","^string_of_channel h
  | `Open_ (n,h) -> "Open_"^n^","^string_of_channel h
let string_of_prequest = function 
  | (r,j,e) -> Printf.sprintf "{[%s]/%d,[%s]}" (string_of_list string_of_request r) j (string_of_list string_of_channel e)
let string_of_completion = function
  | `Go       h -> "Go "     ^string_of_channel h
  | `Gof      h -> "Gof "    ^string_of_channel h
  | `Migrate  h -> "Migrate "^string_of_channel h
  | `OKin     h -> "OKin "   ^string_of_channel h
  | `Register(b,p,s) -> Printf.sprintf "Register %b,<..>" b
let string_of_message = function
  | `R r       -> string_of_prequest r
  | `C c       -> string_of_completion c
  | `Decr      -> "--"
  | `Debug     -> "debug"
  | `Stop      -> "stop"
  | `Step w    -> "step "^w
  | `Eval _    -> "eval .."
  | `Collect _ -> "collect"

open Format
open Printer

let format_channel    = format_string string_of_channel
let format_request    = format_string string_of_request
let format_prequest   = format_string string_of_prequest
let format_completion = format_string string_of_completion
let format_message    = format_string string_of_message

let format_frozen f = function
  | `Out_    (s,p) -> fprintf f "@[out_%s,%a@]"  s format_process p
  | `R_in_   (s,p) -> fprintf f "@[in_%s,%a@]"   s format_process p
  | `R_open_ (s,p) -> fprintf f "@[open_%s,%a@]" s format_process p
  | `Out     (s,c) -> fprintf f "@[out %s,%a@]"  s format_channel c

let format_match f = function
  | `Com   (pl,sl,p) -> fprintf f "@[<%a>,%a@]" format_process_t (Send sl) format_process_t (Rcv (pl,p))
  | `Print (s,p)     -> fprintf f "@[print %s.%a@]" s format_process p
  | `Pause (i,s,p)   -> fprintf f "@[print %i,%s.%a@]" !i s format_process p
  | `Sleep (i,p)     ->	fprintf f "@[sleep %i.%a@]" i format_process p
  | `Req   (r,p)     ->	fprintf f "@[{%a}.%a@]" format_request r format_process p
  | `In    (h,k,n)   ->	fprintf f "@[IN(%a,%a,%s)@]" format_channel h format_channel k n
  | `Out   (h,p)     ->	fprintf f "@[OUT(%a,%a)@]" format_channel h format_process p
  | `Open  (p,h)     -> fprintf f "@[OPEN(%a,%a)@]" format_process p format_channel h

let format_state f s = 
  pp_open_box f 2;
  pp_print_string f s.name;
  pp_print_char f '[';
  if not (Coms. is_empty s.coms)  then fprintf f "coms=%a;@ "  (Coms.format  format_channel) s.coms;
  if not (Outs. is_empty s.outs)  then fprintf f "outs=%a;@ "  (Outs.format  format_channel) s.outs;
  if not (Ins.  is_empty s.ins)   then fprintf f "ins=%a;@ "   (Ins.format   format_channel) s.ins;
  if not (Opens.is_empty s.opens) then fprintf f "opens=%a;@ " (Opens.format format_channel) s.opens;
  if s.frozen<>[]                 then fprintf f "frozen=%a;@ "  (format_list format_frozen) s.frozen;
  if s.matches<>[]                then fprintf f "matches=%a"     (format_list format_match) s.matches;
  pp_print_char f ']';
  pp_close_box f ()

let print_state s = printf "%a@." format_state s

(*
class ['pp] printer p =
object(s)
  inherit ['pp] Printer.printer p

  method print_channel h = 
    pp#print_word (string_of_channel h);
    
  method print_request ((c,n),h) = 
    let c = match c with
      | `In    -> "in "
      | `Out   -> "out "
      | `In_   -> "in_ "
      | `Open_ -> "open_ "
    in
      pp#print_char '{'; pp#print_kwd c; pp#print_word n; 
      s#print_comma; s#print_channel h; pp#print_char '}';
    
  method print_state t = 
    pp#print_word t.name;
    pp#print_char '['; pp#open_box;
    if not (Coms.is_empty t.coms) then (
      pp#newline;
      pp#print_word "coms=";
      List.iter (fun e -> s#print_t (Send e); s#print_semi) 
	(Coms.get_left t.coms);
      List.iter (fun (x,p) -> s#print_t (Rcv(x,p)); s#print_semi)
	(Coms.get_right t.coms);
    );
    if not (Ins.is_empty t.ins) then (
      pp#newline;
      pp#print_word "ins=";
      List.iter (fun (n,c) -> pp#print_kwd "in "; pp#print_word n; s#print_comma; s#print_channel c; s#print_semi) 
	(Ins.get_left t.ins);
      List.iter (fun (n,c) -> pp#print_kwd "in_ "; pp#print_word n; s#print_comma; s#print_channel c; s#print_semi)
	(Ins.get_right t.ins);
    );
    if not (Outs.is_empty t.outs) then (
      pp#newline;
      pp#print_word "outs=";
      List.iter (fun c -> pp#print_kwd "out "; s#print_channel c; s#print_semi) 
	(Outs.get_left t.outs);
      List.iter (fun p -> pp#print_kwd "out_"; s#print_dot; s#print p; s#print_semi) 
	(Outs.get_right t.outs);
    );
    if not (Opens.is_empty t.opens) then (
      pp#newline;
      pp#print_word "opens=";
      List.iter (fun (n,p) -> pp#print_kwd "open "; pp#print_word n; s#print_dot; s#print p; s#print_semi) 
	(Opens.get_left t.opens);
      List.iter (fun (n,c) -> pp#print_kwd "open_ "; pp#print_word n; s#print_comma; s#print_channel c; s#print_semi) 
	(Opens.get_right t.opens);
    );
    if t.frozen<>[] then (
      pp#newline;
      pp#print_word "frozen=";
      List.iter 
	(function
	   | `Out     (n,h) -> 
	       pp#print_kwd "out "; pp#print_word n; s#print_comma; s#print_channel h; s#print_semi
	   | `Out_    (n,p) -> 
	       pp#print_kwd "out_ "; pp#print_word n; s#print_dot; s#print p; s#print_semi
	   | `R_in_   (n,p) -> 
	       pp#print_kwd "in_ "; pp#print_word n; s#print_dot; s#print p; s#print_semi
	   | `R_open_ (n,p) -> 
	       pp#print_kwd "open_ "; pp#print_word n; s#print_dot; s#print p; s#print_semi
	) t.frozen;
    );
    if t.matches<>[] then (
      pp#newline;
      pp#print_word "matches=";
      List.iter 
	(function
	   | `Com   (x,e,p) -> s#print_t (Send e); s#print_t (Rcv(x,p)); s#print_semi
	   | `Print (w,p)   -> pp#print_kwd "print "; pp#print_word w; s#print_dot; s#print p; s#print_semi
	   | `Pause (r,w,p) when !r=0 -> 
	                       pp#print_kwd "pause "; pp#print_word w; s#print_dot; s#print p; s#print_semi
	   | `Pause (r,w,p) -> pp#print_kwd "pause "; pp#print_word (string_of_int !r^" "^w); s#print_dot; s#print p; s#print_semi
	   | `Sleep (i,p)   -> pp#print_kwd "sleep "; pp#print_word (string_of_int i); s#print_dot; s#print p; s#print_semi
	   | `Req   (r,p)   -> s#print_request r; s#print_dot; s#print p; s#print_semi
	   | `In    (h,k)   -> pp#print_kwd "in "; s#print_channel h; s#print_comma; s#print_channel k; s#print_semi
	   | `Out   (h,p)   -> pp#print_kwd "out "; s#print_channel h; s#print_comma; s#print p; s#print_semi
	   | `Open  (p,h)   -> pp#print_kwd "open "; s#print p; s#print_comma; s#print_channel h; s#print_semi
	) t.matches;
    );
    pp#close_box;
    pp#print_char ']';
    pp#newline;

end
*)
