(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Communications of the GC-PAN *) 

open GenComm
open Ast

(* channels *)
type channel

(* located ambient state *)
type state

(* request messages *)
type request = 
    [ `In    of string * channel
    | `In_   of string * channel
    | `Open_ of string * channel
    | `Out   of string * channel
    | `DoOut of channel * channel ] 

(* paqued request messages *)
type prequest = request list * int * channel list

(* completion messages *)
type completion =
    [ `Go       of channel
    | `Gof      of channel
    | `Migrate  of channel
    | `OKin     of channel
    | `Register of bool * process * state ]

(* maintainance messages *)
type local_command =
    [ `Eval of process
    | `Step of string
    | `Collect of Tree.collector
    | `Debug
    | `Stop ]

(* general messages *)
type message = 
    [ `C of completion 
    | `R of prequest
    | `Decr
    | local_command ]

(* command messages *)
type command =
    [ `Create of string * process * channel 
    | `Eval of Ast.process 
    | `Register of channel
    | `Clear
    | `Stop
    | `Step of string
    | `Debug
    | `Tree ]



(* -- Communication functions -- *)

(* send a completion on a channel *)
val send_compl : channel -> completion -> unit

(* send a request on a channel *)
val send_req : channel -> request -> unit

(* forward a paqued request on a channel *)
val send_preq : channel -> prequest -> unit

(* send a `Decr message on a channel *)
val send_decr : channel -> unit

(* send an "eval p" command on a channel *)
val send_eval : channel -> process -> unit

(* broadcast a Gof message *)
val broadcast_gof: channel list -> channel -> unit


(* create a new channel, and listen on it with the given function  *)
val channel : (channel -> message receiver) -> channel


(* send a command to an host (default to the local site) *)
val send_command : ?host:host -> command -> unit

(* broadcast a local command on every local channel;
 * return the number of message sent
 *)
val local_broadcast : local_command -> int

(* create a master for the local site with the given command handler,
 *  listening on the given port
 *)
val master : (command -> unit) -> unit

(* purger for Stopped agents *)
val stop_purge : message receiver

(* collector for the local net *)
val collect: Tree.ops

(* is a channel local *)
val is_local: channel -> bool




(* -- States handling -- *)

(* empty state, with the given name *)
val empty_state: string -> state

(* adding terms to a state 
 * [store s ~create h i p] adds the process p to the state s at channel h, 
 * using the function create to create sub-ambients.
 * return the new state, and the new counter (i)
 *)
val store: state -> create:(?host:host -> ?h:channel -> ?n:string -> process -> channel*int) 
  -> channel -> int -> process -> int * state

(* adding a state to an other one *)
val merge: state -> channel -> state -> state

(* add requests to a state *)
val add_req:  state -> request -> state
val add_reqs: state -> request list -> state


(* get a match from the state *)
val get_match: state -> state *
  [ `None 
  | `Com   of string list * path list * process 
  | `In    of channel * channel * string
  | `Out   of channel * process 
  | `DoOut of channel * channel 
  | `Open  of process * channel
  | `Req   of request * process
  | `Print of string * process
  | `Pause of int ref * string * process
  | `Sleep of int * process ]
  
(* get an async match from the state (for waiting ST) *)
val get_async_match: state -> state * 
  [ `None 
  | `In    of channel * channel * string
  | `DoOut of channel * channel ]


(* get a non "`Req" match from the state (for Immobiles) *)
val get_nonReq_match: state -> state * 
  [ `None 
  | `Com   of string list * path list * process 
  | `In    of channel * channel * string
  | `Out   of channel * process 
  | `DoOut of channel * channel 
  | `Open  of process * channel
  | `Print of string * process
  | `Pause of int ref * string * process
  | `Sleep of int * process ]


(* get the name of a state *)
val state_name: state -> string




(* -- Names handling -- *)

(* create a fresh name with the given base *)
val fresh_name: string -> string

(* this flag controls wether the masterid is included in fresh names or not 
 * (not including it is unsafe when there are several hosts, 
 *  but safe and more readable when we are the only one) *)
val safe_names: bool ref




(* -- Pretty-printing -- *)
val string_of_channel:    channel    -> string
val string_of_request:    request    -> string
val string_of_prequest:   prequest   -> string
val string_of_completion: completion -> string
val string_of_message:    message    -> string

val format_channel:    Format.formatter -> channel    -> unit
val format_request:    Format.formatter -> request    -> unit
val format_prequest:   Format.formatter -> prequest   -> unit
val format_completion: Format.formatter -> completion -> unit
val format_message:    Format.formatter -> message    -> unit
val format_state:      Format.formatter -> state      -> unit

val print_state: state -> unit
