(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* the GC-PAN *)

open Ast
open Comm


let cf c h k   = c#forwarder (string_of_channel h) (string_of_channel k)
let ca c h s k = c#ambient (string_of_channel h) (state_name s) (string_of_channel k)
let weight n = if immobile n then 2 else 1

(* forwarders relocation mode *)
let relocation_mode = ref `Mixed
(* pack requests when unblocking a forwarder *)
let pack_requests = ref true 

(* purger for GCed/Stopped agents *)
let gc_purge = 
  `Close
    (function
       | `Stop      -> ()
       | `Collect o -> o#dead
       | `Debug     -> Tree.debug "purged agent"
       | _          -> failwith "GC agent received a message")

(* code for the forwarders 
 * (we need to record the "non-received" requests)
 *)
let rec forwarder ?(q=[]) h i k = 
  let add h e =
    let rec aux (l,e) = function
      | []                   -> l,h::e
      | x::q when is_local x -> aux (x::l,e) q
      | x::q                 -> aux (l,x::e) q
    in match !relocation_mode with
      | `Sync  -> e, [h]
      | `Async -> h::e, []
      | `Mixed -> aux ([],[]) e
  in
  let handle_req q ?(o=1) (req,j,e) = 
    if i-o < 0 then (send_preq k (req,j,e); Stats.gc_forwarder(); gc_purge)
    else let b,s = add h e in
      send_preq k (req, j+1, s); bforwarder ~q h (i-o) k b
  in match q with 
    | [] ->
	`Receive 
	  (fun m -> match m with
	     | `R r           -> handle_req [] r
	     | `Stop          -> stop_purge
	     | `Decr          -> 
		 if i>0 then forwarder h (i-1) k
		 else (send_decr k; Stats.gc_forwarder(); gc_purge)
	     | `Step _        -> forwarder h i k
	     | `Collect c     -> cf c h k; forwarder h i k
	     | `Debug         -> 
		 Tree.debug "%s > %s ^ %i" (string_of_channel h) (string_of_channel k) i;
		 forwarder h i k
	     | `C c           -> failwith "forwarder received a completion: `%a'" !!format_completion c
	     | #local_command -> failwith "forwarder received a illegal local command"
	  )
    | q when !pack_requests -> 
	let r,o = List.fold_left 
			(fun ((req,j,e),o) (req',j',e') -> 
			   (List.rev_append req' req, j'+j, List.rev_append e' e), o+1)
			(([],0,[]),0)
			q
	in
	  handle_req [] ~o r
    | r::q -> handle_req q r 

(* code for the blocked forwarders 
 * (k is here only for pretty-printing)
 *)
and bforwarder ?(q=[]) h i k e =
  `Receive
    (function
       | `C (`Gof k)    -> broadcast_gof e k; forwarder ~q:(List.rev q) h i k
       | `R req         -> bforwarder ~q:(req::q) h i k e
       | `Decr          -> bforwarder ~q h (i-1) k e
       | `Step _        -> bforwarder ~q h i k e
       | `Stop          -> stop_purge
       | `Collect c     -> cf c h k; bforwarder ~q h i k e
       | `Debug         -> 
	   Tree.debug "%s < %s ^ %i" (string_of_channel h) (string_of_channel k) i;
	   bforwarder ~q h i k e
       | `C c           -> failwith "blocked forwarder received a completion: `%a'" !!format_completion c
       | #local_command -> failwith "blocked forwarder received a illegal local command"
    )

(* out forwarders *)
and out_forwarder k i = 
  Stats.out_forwarder i; 
  channel (fun h -> forwarder h (i-1) k)


(* code for localized ST ambients *)
and st_ambient h i k s p =
  (* create sub-ambients,
   * add the process p to the local process, 
   * search for a potential match,
   * execute it or wait
   *)
  let i,s = store ~create s h i p in
  let s,m = get_match s in
    match m with
      | `None               -> st_ambient_rcv h i k s
      | `Com (xl, paths, p) -> Stats.l_com(); st_ambient h i k s (Subst.names xl paths p)
      | `In  (h', k', n)    ->  
	  Stats.l_in();
	  send_compl h' (`Go   k');
	  send_compl k' (`OKin h);
	  st_ambient h (i+weight n) k s zero
      | `Out (h', p)        -> 
	  Stats.l_out();
	  let k' = out_forwarder k 2 in 
	    send_compl h' (`Go k');
	    st_ambient h i k' s p  
      | `DoOut (h1, h2)     -> 
	  send_compl h1 (`Go h);
	  send_compl h2 (`Go h);
	  st_ambient h (i+2) k s p  (* i+1+weight n *)
      | `Open (p, h')       -> 
	  Stats.l_open();
	  send_compl h' (`Migrate h);
	  wait_st_ambient h (i+1) k s p  
      | `Req (r, p)         -> 
	  send_req k r;
	  wait_st_ambient h i k s p  
      | `Print (w, p)       -> 
	  print_string w; print_newline();
	  st_ambient h i k s p  
      | `Pause (r, w, p) when !r=0 -> wait_st_ambient h i k s ~w p  
      | `Pause (r, w, p)    -> decr r; st_ambient h i k s p  
      | `Sleep (t, p)       -> 
	  Thread.delay (float_of_int t);
	  st_ambient h i k s p  

(* reception of requests *)
and st_ambient_rcv h i k s =
  `Receive
    (function
       | `R (req, j, e) -> 
	   Stats.request (j, List.length req);
	   broadcast_gof e h;
	   st_ambient h (i+j-1) k (add_reqs s req) zero
       | `Decr          -> st_ambient_rcv h (i-1) k s
       | `Step _        -> st_ambient_rcv h i k s
       | `Stop          -> stop_purge
       | `Collect c     -> ca c h s k; st_ambient_rcv h i k s
       | `Debug         -> 
	   Tree.debug "%s _ %s ^ %i : %a" (string_of_channel h) (string_of_channel k) i !!format_state s;
	   st_ambient_rcv h i k s
       | `Eval p        -> st_ambient h i k s p
       | `C c           -> failwith "non-wait ambient received completion : `%a'" !!format_completion c
    )


and wait_st_ambient h i k s ?w p =
  (* handle existing matches (ins),
   * execute them and wait *)
  match get_async_match s with
    | s, `In (h', k', n) -> 
	Stats.l_in();
	send_compl h' (`Go   k');
	send_compl k' (`OKin h);
	wait_st_ambient h (i+weight n) k s ?w p
    | s, `DoOut (h1, h2)     -> 
	send_compl h1 (`Go h);
	send_compl h2 (`Go h);
	wait_st_ambient h (i+2) k s ?w p  (* i+1+weight n *)
    | s, `None -> 
	wait_st_ambient_rcv h i k s ?w p

(* reception of messages, in wait state (wait.p) *)
and wait_st_ambient_rcv h i k s ?w p =
  `Receive
    (function
       | `R (req, j, e)   -> 
	   Stats.request (j, List.length req);
	   broadcast_gof e h;
	   wait_st_ambient ?w h (i+j-1) k (add_reqs s req) p
       | `C (`Go k')      -> Stats.completion(); st_ambient h i k' s p
       | `C (`OKin k')    -> Stats.completion(); st_ambient h (i+1) k' s p
       | `C (`Migrate k') ->
	   Stats.completion(); 
	   send_compl k' (`Register (i=0, p, s)); 
	   if i=0 then gc_purge
	   else (Stats.forwarder i; forwarder h (i-1) k')
       | `C (`Register (b, q, s')) ->
	   Stats.completion(); 
	   st_ambient h (if b then i-1 else i) k (merge s h s') (par p q)
       | `Decr            -> wait_st_ambient_rcv h (i-1) k s ?w p
       | `Stop            -> stop_purge
       | `Step w'         -> 
	   (match w with
	      | Some w when w=w' || w'="*" -> st_ambient h i k s p
	      | _                          -> wait_st_ambient h i k s ?w p)
       | `Collect c       -> ca c h s k; wait_st_ambient_rcv h i k s ?w p
       | `Debug           -> 
	   Tree.debug "%s _ %s ^ %i : %a@ wait.%a" (string_of_channel h) (string_of_channel k) i !!format_state s !!Printer.format_process p;
	   wait_st_ambient_rcv h i k s ?w p
       | `C _             -> failwith "wait ambient received a Gof"	 
       | #local_command   -> failwith "wait ambient received an illegal local command"
    )


(* creation of an ambient, named n, with father h (or itself by default) or host.root, 
 * with content p *)
and create ?host ?h ?(n=rootname) p = 
  let parent f k = match host with
    | None      -> f k (oget h k)
    | Some host -> 
	send_command ~host (`Register k);
	`Receive (function `C (`Go h) -> f k h 
		    | m -> failwith "bad wait response: `%a'" !!format_message m) 
  in
  if immobile n then
    channel (parent (fun k h -> Immobile.ambient ~gcpan:out_forwarder ~create k h (empty_state n) p)), 2
  else 
    channel (parent (fun k h -> st_ambient k 0 h (empty_state n) p)), 1
