(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Generic communications between Threads and Processes *)


(* type of collections *)
module type COLLECTION = 
sig
  exception Empty
  type 'a t 
  val create: unit -> 'a t
  val get: 'a t -> 'a
  val iter: ('a -> unit) -> 'a t -> unit
  val add: 'a t -> 'a -> unit
  val is_empty: 'a t -> bool
end 

(* Last In Last Out collections *)
module LiFo: COLLECTION = 
struct
  exception Empty
  type 'a t = 'a list ref
  let create () = ref []
  let get l = match !l with
    | [] -> raise Empty
    | x::q -> l := q; x
  let add l x = l := x :: !l
  let iter f l = List.iter f !l
  let is_empty l = !l=[]
end

(* First In First Out collections *)
module FiFo: COLLECTION =
struct
  exception Empty
  type 'a t = ('a list * 'a list) ref
  let create () = ref ([],[])
  let rec get c = match !c with
    | [],[]  -> raise Empty
    | l,x::q -> c := l,q; x
    | l,[]   -> c := [], List.rev l; get c
  let add c x = c := x :: fst !c, snd !c
  let iter f c = List.iter f (fst !c); List.iter f (snd !c)
  let is_empty c = !c = ([],[])
end


(* Threadsafe hashtable *)
module LockedTable(Names: Names.T)(M: sig type t end): sig
  val add: (Names.t -> M.t) -> Names.t * M.t
  val get: Names.t -> M.t
  val iter: (Names.t -> M.t -> unit) -> unit
  val fold: (Names.t -> M.t -> 'a -> 'a) -> 'a -> 'a
  val rem: Names.t -> unit
end = struct
  let t = Hashtbl.create 7
  let m = Mutex.create() 
  let locked f x = 
    Mutex.lock m;
    try let y = f x in Mutex.unlock m; y
    with e -> Mutex.unlock m ; raise e
  let add = 
    locked 
      (fun f -> 
	 let id = Names.fresh() in
	 let x = f id in
	   Hashtbl.add t id x;
	   id,x)
  let rem id = 
    locked 
      (fun () -> 
	 Names.free id;
	 Hashtbl.remove t id
      ) ()
  let get    = locked (Hashtbl.find t)
  let iter f = locked (fun () -> Hashtbl.iter f t) ()
  let fold f = locked (Hashtbl.fold f t)
end



(* Local channels : communication by shared memory *)
module LocalChannel: sig
  (* type of channels *)
  type 'a t
  exception Closed   

  (* creation of a channel *)
  val create: unit -> 'a t

  (* reception on a channel, raise Closed if the channel has been closed *)
  val rcv: 'a t -> 'a

  (* emission on a channel, raise Closed if the channel has been closed *)
  val send: 'a t -> 'a -> unit

  (* close a channel, and purge the remaining messages with the given function,
   * raise Closed if the channel has been closed *)
  val close: 'a t -> ('a -> unit) -> unit
end = struct
  type 'a t = 'a FiFo.t * Mutex.t * Condition.t * bool ref
  exception Closed
  let lock m b = 
    Mutex.lock m;
    if !b then (Mutex.unlock m; raise Closed)
  let create() = 
    FiFo.create(),
    Mutex.create(),
    Condition.create(),
    ref false
  let rcv (s,m,c,b) =
    lock m b;
    while FiFo.is_empty s do
      Condition.wait c m
    done;
    let x = FiFo.get s in
      Mutex.unlock m;
      x
  let send (s,m,c,b) x =
    lock m b;
    FiFo.add s x;
    Condition.signal c;
    Mutex.unlock m
  let close (s,m,_,b) f =
    lock m b;
    b := true;
    try FiFo.iter f s; Mutex.unlock m
    with e -> Mutex.unlock m; raise e
end



(* type of an host (process) : hostname + port number *)
type host = string * int
let string_of_host = function
  | (h,1550) -> h
  | (h,p)    -> h^":"^(string_of_int p)


(* user friendly printing of error *)
let string_of_error = function
  | Unix.Unix_error(e,s,"") -> "Unix error: "^s^" ("^Unix.error_message e^")"
  | Unix.Unix_error(e,s,a ) -> "Unix error: "^s^" "^a^" ("^Unix.error_message e^")"
  | e -> Printexc.to_string e



(* Generic channels : communication by sockets
 * Connections are cached, in order to reduce their number *)
module GenChannel: sig
  (* send a value to an host *)
  val send: host -> 'a -> unit
end = struct
  let t = Hashtbl.create 5
  let lock = Mutex.create()
  let connect (h, p) = 
    let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
    let host = Unix.gethostbyname h in
    let addr = host.Unix.h_addr_list.(0)
    in
      try 
(* 	Printf.printf "connect to %s:%i\n%!" (Unix.string_of_inet_addr addr) p; *)
	Unix.connect sock (Unix.ADDR_INET (addr, p));
	Unix.out_channel_of_descr sock
      with e -> 
	Unix.close sock; 
	match e with 
	  | Unix.Unix_error(e,f,"") -> raise (Unix.Unix_error(e,f,h^":"^string_of_int p))
	  | _ -> raise e
(* tentative d'utilisation du GC ocaml pour fermer les connexions inutiles *)
(* 
  let close h = 
    Printf.printf "GC: closing connection to %s#%i\n%!" 
      (fst h) (snd h);
    let o = Hashtbl.find t h in
      Hashtbl.remove t h;
      close_out o
*)
  let send host = 
    let o =
      Mutex.lock lock;
      let o = 
	try Hashtbl.find t host
	with 
	  | Not_found ->
	      let o = connect host in
		Hashtbl.add t host o;
(* 	          Gc.finalise close host; *)
		o
	  | e -> Mutex.unlock lock; raise e
      in       
	Mutex.unlock lock; o
    in
      fun m -> output_value o m; flush o
end


let localhost = Unix.gethostname()


(* type of a message receiver : 
 *  - handle the message and wait for another one
 *  - close the channel, purge all remaining messages with the given function
 *)
type 'a receiver = [
| `Receive of ('a -> 'a receiver) 
| `Close   of ('a -> unit)
]     
let loop_receiver f = 
  let rec aux x = f x; `Receive aux in
    `Receive aux


(* input signature for the following functor *)
module type CONV = 
sig
  (* type of messages ('c will be the type of channels) *)
  type +'c t
  (* serialization process requires to handle channels, 
   * so that we need a convertion function *)
  val conv: ('c -> 'd) -> 'c t -> 'd t
end



(* Build communication primitives for the given message and command types *)
module Make(Msg:CONV)(Cmd:CONV) =
struct

  (* port of the local master *)
  let port = 
    let p = ref 1550 in
      for i=1 to Array.length Sys.argv - 2 do
	if Sys.argv.(i)="-p" then 
	  try p:=int_of_string Sys.argv.(i+1)
	  with Failure _ -> ()
      done;
      !p
  let localport=localhost,port

  (* unique identifier for the local master *)
  let masterid = "@"^string_of_host localport^"#"


  (* Name generator for channels *)
  module Loc = Names.Make(struct let base="&" end)


  (* type of "generic", marshalable channels *)
  type gchannel = [
  | `Gen of host * Loc.t
  ]

  (* type of channels : with local channels *)
  type channel = [
  | gchannel
  | `Loc of channel Msg.t LocalChannel.t * Loc.t
  ]

  (* table recording all opened local channels *)
  module Chan = LockedTable(Loc)
    (struct type t = channel Msg.t LocalChannel.t * Loc.t end)

  (* type of messages *)
  type message  = channel  Msg.t
  (* type of marshalable messages *)
  type gmessage = gchannel Msg.t

  (* type of commands *)
  type command  = channel  Cmd.t
  (* type of marshalable commands *)
  type gcommand = gchannel Cmd.t

  (* type of messages emitted on the network : marshalable messages and commands *)
  type network_message =
    | Com of gcommand
    | Msg of Loc.t * gmessage


  (* pre-marshaling of a channel value *)
  let gchannel_of_channel = function
    | #gchannel as g -> g
    | `Loc (_,loc) -> `Gen (localport, loc)
  (* pre-marshaling of a message *)
  let message_conv = Msg.conv gchannel_of_channel
  (* pre-marshaling of a command *)
  let command_conv = Cmd.conv gchannel_of_channel


  (* pretty-printing of channels *)
  let string_of_channel = function
    | `Gen (host,loc) when host=localport
	              -> Loc.to_string loc
    | `Gen (host,loc) -> string_of_host host^"#"^Loc.to_string loc
    | `Loc (_,loc)    -> Loc.to_string loc

  (* equality of channels *)
  let eq_channel a b = string_of_channel a = string_of_channel b

  (* is a channel local *)
  let is_local = function
    | `Loc _ -> true
    | `Gen (host,_) -> host=localport


  (* shorthands *)
  let message m loc = Msg (loc, message_conv m)
  let command m     = Com (command_conv m)
  let send_loc loc = 
    let chan,_ = Chan.get loc in
      LocalChannel.send chan

  (* send a message on a channel : depends on the channel kind *)
  let send c m = match c with 
    | `Loc (chan, loc) -> Stats.local_message(); LocalChannel.send chan m
    | `Gen (host, loc) when host=localport
	               -> Stats.local_message(); send_loc loc m
    | `Gen (host, loc) -> Stats.net_message(); GenChannel.send host (message m loc)


  (* broadcast a message to every local channel, and return the number of "success" *)
  let local_broadcast m =
    Chan.fold 
      (fun _ (c,_) n -> 
	 try LocalChannel.send c m; n+1
	 with LocalChannel.Closed -> n
      ) 0


  (* setup of a channel receiver *)
  let channel f =
    let lchan = LocalChannel.create() in
    let loc, chan = Chan.add (fun id -> lchan, id) in
    let rec listen = function
      | `Receive f -> listen (f (LocalChannel.rcv lchan))
      | `Close   g -> Chan.rem loc; LocalChannel.close lchan g
    in
    let slisten m = 
      try listen m with e -> 
	Chan.rem loc; 
	LocalChannel.close lchan (fun _->()); 
	Printf.printf "Channel %s rose exception `%s'\n%!"
	  (string_of_channel (`Loc chan))
	  (string_of_error e)
    in
      Thread.create slisten (f (`Loc chan)), `Loc chan


  (* send a command to an host *)
  let send_command ?(host=localport) m =
    GenChannel.send host (command m)    



  (* setup of an internet "listener" on the given port *)
  let master f =
    let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
    let addr = Unix.ADDR_INET (Unix.inet_addr_any,port) in
    let listen s = 
      try while true do
	match input_value s with
	  | Com m -> f (m:>command)
	  | Msg (loc,m) -> send_loc loc (m:>message) 
      done with End_of_file -> ()
    in
    let rec server () =
      let s, caller = Unix.accept sock in
	ignore(Thread.create listen (Unix.in_channel_of_descr s));
	server()
    in
      Unix.setsockopt sock Unix.SO_REUSEADDR true;
      Unix.bind sock addr;
      Unix.listen sock 5;
      Thread.create server ()

end
