(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Generic communications between Threads and Processes *)


(* type of an host (process) : hostname + port number *)
type host = string*int
val string_of_host: host -> string

(* type of a message receiver : 
 *  - handle the message and wait for another one
 *  - close the channel, purge all remaining messages with the given function
 *)
type 'a receiver = [
| `Receive of ('a -> 'a receiver) 
| `Close   of ('a -> unit)
]     
val loop_receiver: ('a -> unit) -> 'a receiver


(* input signature for the following functor *)
module type CONV = 
sig
  (* type of messages ('c will be the type of channels) *)
  type +'c t
  (* serialization process requires to handle channels, 
   * so that we need a convertion function *)
  val conv: ('c -> 'd) -> 'c t -> 'd t
end


(* Build communication primitives for the given message and command types *)
module Make(Msg:CONV)(Cmd:CONV):
sig

  (* type of communication channels *)
  type channel

  (* type of messages *)
  type message = channel Msg.t

  (* type of commands *)
  type command = channel Cmd.t
    

  (* send a message on a channel *)
  val send: channel -> message -> unit

  (* [channel f] setup a new channel c with the receiver (f c)
   * return the corresponding thread uid, and c *) 
  val channel: (channel -> message receiver) -> Thread.t*channel

  (* broadcast a message on every channel local to the process,
   * return the number of channel reached *)
  val local_broadcast: message -> int
    

  (* send a command to host (the local process by default) *)
  val send_command: ?host:host -> command -> unit

  (* [master f] setup an internet "listener" on localhost, 
   * using the function f.
   * return the corresponding thread uid*)
  val master: (command -> unit) -> Thread.t


  (* pretty-printing of channels *)
  val string_of_channel: channel -> string

  (* equality on channels *)
  val eq_channel: channel -> channel -> bool

  (* is a channel local *)
  val is_local: channel -> bool

  (* string uniquely identifying the local master (listener) *)
  val masterid: string

end
