(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Code for immobile ambients *)

open Ast
open Comm


(* remove pauses from an association list *)
let rem_pauses_assoc x = 
  let rec aux (acc,rem) = function
    | [] -> acc,rem
    | (y,v)::q when x="*" || x=y -> aux (par v acc,rem) q
    | x::q -> aux (acc,x::rem) q
  in aux (zero,[])

(* out=
 0: waiting for OKin (-> prochain out direct)
 2: waiting for OKin, an out done (-> prochain out chapeaut 2)
 3: not waiting (-> prochain out chapeaut 3)
*)

let ambient ?gcpan ~(create:?host:GenComm.host -> ?h:channel -> ?n:string -> process -> channel*int) = 

  let handle_match amb amb_rcv out wl h k s p = function
    | `None               -> amb_rcv out wl h k s
    | `Com (xl, paths, p) -> Stats.l_com(); amb out wl h k s (Subst.names xl paths p)
    | `In  (h', k', _)    ->  
	Stats.l_in(); 
	send_compl h' (`Go   k');
	send_compl k' (`OKin h );
	amb out wl h k s zero
    | `Out (h', p)        -> 
	Stats.l_out(); 
	(match gcpan,out with
	   | None,_ | _,0 -> 
	       send_compl h' (`Go k);
	       amb 2 wl h k s p
	   | Some fwd,_ ->
	       let k' = fwd k out in
		 send_compl h' (`Go k');
		 amb out wl h k' s p  
	)
    | `DoOut (h1, h2)     -> 
	send_compl h1 (`Go h);
	send_compl h2 (`Go h);
	amb out wl h k s p
    | `Open (p, h')       -> 
	Stats.l_open(); 
	send_compl h' (`Migrate h);
	amb out wl h k s p
    | `Req (_, p) -> failwith "Immobile ambient wants to move"
    | `Print (w, p)       -> 
	print_string w; print_newline();
	amb out wl h k s p
    | `Pause (r, w, p) when !r=0 -> amb out ((w,p)::wl) h k s zero
    | `Pause (r, w, p)    -> decr r; amb out wl h k s p
    | `Sleep (t, p)       -> 
	Thread.delay (float_of_int t);
	amb out wl h k s p 
  in
  let handle_msg amb amb_rcv = fun out wl h k s -> function
    | `R (req, j, e) -> 
	Stats.request (j, List.length req); 
	broadcast_gof e h;
	amb out wl h k (add_reqs s req) zero
    | `C (`Register (_, p, s')) ->
	Stats.completion(); 
	amb out wl h k (merge s h s') p
    | `Decr          -> amb_rcv out wl h k s

    | `C c           -> failwith "Immobile ambient received a bad completion : `%a'" !!format_completion c
    | `Step w        -> 
	let p,wl = rem_pauses_assoc w wl in
	  amb out wl h k s p
    | `Collect c     -> 
	c#ambient (string_of_channel h) (state_name s) (string_of_channel k); 
	amb_rcv out wl h k s
    | `Debug         -> 
	Tree.debug "%s _ %s : %a" (string_of_channel h) (string_of_channel k) !!format_state s;
	amb_rcv out wl h k s
    | `Stop          -> stop_purge
    | `Eval p        -> amb out wl h k s p
  in

  (* code for immobile ambients *)
  let rec ambient out wl h k s p =
    let _,s = store ~create s h 0 p in
    let s,m = get_match s in
      match m with
	| `Req (`In_ _ as r, p) -> 
	    send_req k r;
	    ambient' p 0 wl h k s zero
	| _ -> handle_match ambient ambient_rcv out wl h k s p m

  (* reception for immobile ambients *)
  and ambient_rcv out wl h k s = `Receive (handle_msg ambient ambient_rcv out wl h k s)

  (* immobile ambient which has emitted an in_ request
   * (it cannot emit another one until the former has been completed)
   *)
  and ambient' q = fun out wl h k s p ->
    let _,s = store ~create s h 0 p in
    let s,m = get_nonReq_match s in 
      handle_match (ambient' q) (ambient'_rcv q) out wl h k s p m

  (* reception for immobile ambients which have emitted an in_ request *)
  and ambient'_rcv q = fun out wl h k s ->
    `Receive
      (function
	 | `C (`OKin k') ->
	     Stats.completion(); 
	     if gcpan=None then ambient out wl h k s q
	     else (
	       if out=2 then send_decr k;
	       ambient 3 wl h k' s q
	     )
	 | `Debug -> 
	     Tree.debug "%s _ %s : %a@ wait.%a" 
	       (string_of_channel h) (string_of_channel k) !!format_state s !!Printer.format_process q;
	     ambient'_rcv q out wl h k s
	 | m -> handle_msg (ambient' q) (ambient'_rcv q) out wl h k s m
      )

  in
    ambient 3 []

