(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Lexing *)

{ 
  open Ast 
  open Lexing

  let fail p m = 
    raise (ParseError (p.pos_lnum, p.pos_cnum,m))

  let newline lexbuf =
    let pos = lexbuf.lex_curr_p in
    lexbuf.lex_curr_p <- 
      { pos with pos_lnum = pos.pos_lnum + 1; pos_bol = pos.pos_cnum }

}

let digit    = [ '0'-'9' ]
let minusc   = [ 'a'-'z' '' '' '' '' '' '' '' '' '' '' '' '' '']
let majusc   = [ 'A'-'Z' '' '' '' '' '' '' '' '' '' '' '' '' '']
let blank    = [' ' '\r' '\t']
let alpha    = minusc | majusc
let alphanum = alpha | digit [ '-' '\'' ]
let word     = alpha alphanum*
let host     = word ('.' word)* 

rule token = parse
  | '\n'           { newline lexbuf; token lexbuf }
  | "(*"           { comment lexbuf; token lexbuf }
  | blank*         {                 token lexbuf }
  | ";;"           { SEMI  }
  | '='            { EQUAL }
  | '\''           { QUOTE }
  | ':'            { COLON }
  | ','            { COMMA }
  | '.'            { DOT   }
  | '('            { LPAR  }
  | ')'            { RPAR  }
  | '['            { LSQR  }
  | ']'            { RSQR  }
  | '<'            { LPIK  }
  | '>'            { RPIK  }
  | '|'            { BAR   }
  | "nu"           { NU    }
  | "rec"          { REC   }
  | "sleep"        { SLEEP }
  | '!'            { BANG  }
  | '*'            { STAR  }
  | '0'            { ZERO  }
  | "let"          { LET   }
  | "in"           { CAPA In    }
  | "out"          { CAPA Out   }
  | "open"         { CAPA Open  }
  | "in_"          { CAPA In_   }
  | "out_"         { CAPA Out_  }
  | "open_"        { CAPA Open_ }
  | "print"        { CAPA Print }
  | "pause"        { CAPA (Pause (ref 0)) }

  | "#quit" | eof  { EOF }
  | '#'(word as w) { DIESE w }
  | '@'(host as h) { HOST  h }
  | '~' word as w  { WORD  w }
  | word as w      { WORD  w }
  | digit+ as s    { INT (int_of_string s) }
  | _ as s         { error (lexeme_start_p lexbuf) 
		      ("Unrecognized character `"^String.make 1 s^"'") lexbuf }

and comment = parse
  | '\n'           { newline lexbuf; comment lexbuf }
  | "(*"           { comment lexbuf; comment lexbuf }
  | "*)"           { ()                             } 
  | eof            { fail (lexeme_start_p lexbuf) "Unterminated comment" }
  | _              { comment lexbuf                 }

and error p m = parse
  | '\n'           { newline lexbuf; error p m lexbuf }
  | "(*"           { (try comment lexbuf with ParseError _ -> ()); error p m lexbuf }
  | ";;" | eof     { fail p m }
  | _              { error p m lexbuf }

