(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* main entry *)

open Ast
open Comm

let create = ref GcPan.create
let m c () = create := c

let () = 
  let pan() = create := Pan.create in
  let usage = "usage: gcpan [-p port] [-pan] [other options...]" in
  let rec specs =
      [ "-p",     Arg.Int  (fun _->()(*set in GenComm*)), "                       : set the port (default to 1550)";
	"-pan",   Arg.Unit (m Pan.create), "                     : PAN mode (no gc)";
	"-pack",  Arg.Symbol (["on"; "off"], fun s -> GcPan.pack_requests:=(s="on")), "           : GCPAN requests packing";
	"-reloc", Arg.Symbol (["sync"; "async"; "mixed"], 
			      function
				| "sync"  -> GcPan.relocation_mode:=`Sync
				| "async" -> GcPan.relocation_mode:=`Async
				| _       -> GcPan.relocation_mode:=`Mixed), ": GCPAN forwarders relocation mode";
	"-version", Arg.Unit (fun _->Parser.print_version(); exit 0), "                 : print version and licence informations";
	"-help",    Arg.Unit (fun _->Parser.print_help(); exit 0), "                    : print some help on the interpreter";
	"--help",   Arg.Unit (fun _->Arg.usage specs usage; exit 0), "                   : print this list of options";
      ]
  in
    Arg.parse specs (fun s -> raise (Arg.Bad ("unknown argument : "^s))) usage

let _ = 
  let root = ref (fst (!create zero)) in
    master 
      (function 
	 | `Create (n,p,h) -> ignore (!create ~h ~n p)
	 | `Register h     -> send_compl h (`Go !root) 
	 | `Eval    p      -> send_eval !root p
	 | `Clear          -> ignore(local_broadcast `Stop); root := fst (!create zero)
	 | `Stop           -> exit 0
	 | `Step w         -> ignore(local_broadcast (`Step w))
	 | `Debug          -> collect#debug
	 | `Tree           -> collect#print_ftree
      )

let _ = 
  let add p = 
    try 
      ignore(Check.free_names p);
      send_command (`Eval p);
    with Failure s -> print_endline s
  in
  let addto (host,n,p) = 
    try 
      ignore(Check.free_names p);
      ignore(!create ~host ~n p);
    with Failure s -> print_endline s
  in
  let s = Parser.interactive stdin in
    Stream.iter 
      (function
	 | `Stats       -> Stats.print()
	 | `SafeNames b -> safe_names := b; Printf.printf "Safe Names turned %s.\n%!" (if b then "on" else "off")
	 | `PackReq   b -> GcPan.pack_requests := b; Printf.printf "Requests packing turned %s.\n%!" (if b then "on" else "off")
	 | `RelocMode b -> GcPan.relocation_mode := b; 
	     (match b with
		| `Sync  -> Printf.printf "Synchronous forwarders relocation\n%!" 
		| `Async -> Printf.printf "Asynchronous forwarders relocation\n%!" 
		| `Mixed -> Printf.printf "Mixed forwarders relocation\n%!" )
	 | `Cmd c       -> send_command (c:>command)
	 | `Def f       -> Printf.printf "`%s' is defined.\n%!" f
	 | `Print p     -> Printer.print_process p
	 | `Check p     -> Printf.printf "%i type(s) inferred.\n%!" 
	     (List.length (Type.infer (Hashtbl.create 5,Hashtbl.create 5) ([],[],[],[]) p))
	 | `Add p       -> add p
	 | `Eval p      -> send_command `Clear; Stats.reset(); add p
	 | `AddTo x     -> addto x
	 | `Help        -> Parser.print_help()
      ) s;
    ignore(local_broadcast `Stop);
    print_endline "ciao..."
