(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Generic bi-matcher *)

type 'c formatter = (Format.formatter -> 'c -> unit)

module type CONV = sig
  include GenComm.CONV
  val format: 'c formatter -> 'c t formatter
end

module type MATCHER = 
sig
  type 'c l
  type 'c r
  type 'c lr
  include CONV
  type 'c tlr = 'c t * 'c lr list
  val add_left : 'c tlr -> 'c l -> 'c tlr
  val add_right: 'c tlr -> 'c r -> 'c tlr
  val add:       'c tlr -> 'c t -> 'c tlr
  val get_left:  'c t -> 'c l list
  val get_right: 'c t -> 'c r list
  val empty:     'c t
  val is_empty:  'c t -> bool
end

let format_list format_elt = 
  let rec aux f = function
    | [] -> ()
    | [x] -> format_elt f x
    | x::q -> Format.fprintf f "%a,@ %a" format_elt x aux q
  in aux

module Simple(L: CONV)(R: CONV)(U: sig type 'c t val mix: 'c L.t -> 'c R.t -> 'c t end) =
struct

  type 'c l  = 'c L.t
  type 'c r  = 'c R.t
  type 'c lr = 'c U.t
  type 'c t = 
    | L of 'c l list
    | R of 'c r list
  type 'c tlr = 'c t * 'c lr list

  let empty = L []

  let conv f = function
    | L l -> L (List.map (L.conv f) l)
    | R r -> R (List.map (R.conv f) r)

  let add_left (m,l) x = match m with
    | L q      -> L (x::q), l
    | R []     -> L [x], l
    | R (y::q) -> R q, U.mix x y :: l

  let add_right (m,l) y = match m with
    | R q      -> R (y::q), l
    | L []     -> R [y], l
    | L (x::q) -> L q, U.mix x y :: l

  let add = 
    let rec aux acc = function
      | [], r -> R r, acc
      | l, [] -> L l, acc
      | x::qx, y::qy -> aux (U.mix x y :: acc) (qx,qy)
    in
      fun (m,q) m' -> match m,m' with
	| L l, L l' -> L (List.rev_append l l'), q
	| R r, R r' -> R (List.rev_append r r'), q
	| L l, R r
	| R r, L l  -> aux q (l,r)

  let get_left  = function L l -> l | _ -> []
  let get_right = function R l -> l | _ -> []

  let is_empty = function 
    | L [] | R [] -> true | _ -> false

  let format g f = function
    | L l -> Format.fprintf f "L[@[%a@]]" (format_list (L.format g)) l
    | R r -> Format.fprintf f "R[@[%a@]]" (format_list (R.format g)) r

end


module Named(M: MATCHER) =
struct

  type 'c t = (string * 'c M.t) list
  type 'c l = string * 'c M.l
  type 'c r = string * 'c M.r
  type 'c lr  = 'c M.lr
  type 'c tlr = 'c t * 'c lr list

  let empty = []

  let conv f =
    List.fold_left (fun acc (n,m) -> (n,M.conv f m) :: acc) []


  let rem n = 
    let rec aux acc =
      function
	| [] -> M.empty, acc
	| (n',m)::q when n=n' -> m, List.rev_append acc q
	| x::q -> aux (x::acc) q
    in aux []

  let add_left (t,l) (n,x) = 
    let m,t = rem n t in 
    let m,l = M.add_left (m,l) x in
      if M.is_empty m then t,l else (n,m)::t,l

  let add_right (t,l) (n,y) =
    let m,t = rem n t in 
    let m,l = M.add_right (m,l) y in
      if M.is_empty m then t,l else (n,m)::t,l

  let rec add x = 
    List.fold_left 
      (fun (t,l) (n,m') -> 
	 let m,t = rem n t in
	 let m,l = M.add (m,l) m' in
	   (n,m)::t, l) x

  let get_left t = 
    List.fold_left (fun acc (n,m) -> 
		      List.fold_left (fun acc x -> (n,x)::acc) acc (M.get_left m)
		   ) [] t

  let get_right t =
    List.fold_left (fun acc (n,m) -> 
		      List.fold_left (fun acc x -> (n,x)::acc) acc (M.get_right m)
		   ) [] t

  let is_empty t = t=[]

  let format g = format_list (fun f (n,m) -> Format.fprintf f "@[%s@ ->@ %a@]" n (M.format g) m)

end


module SimpleNamed(L: CONV)(R: CONV)(U: sig type 'c t val mix: 'c L.t -> 'c R.t -> 'c t end) =
  Named(Simple(L)(R)(U))
