(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* the PAN *)

open Ast
open Comm

let cf c h k   = c#forwarder (string_of_channel h) (string_of_channel k)
let ca c h s k = c#ambient (string_of_channel h) (state_name s) (string_of_channel k)

(* code for the forwarders *)
let rec forwarder h k = 
  `Receive 
    (fun m -> match m with
       | `R (r,j,e)     -> send_preq k (r,j+1,e); forwarder h k
       | `C c           -> send_compl k c; forwarder h k
       | `Decr
       | `Step _        -> forwarder h k
       | `Stop          -> stop_purge
       | `Collect c     -> cf c h k; forwarder h k
       | `Debug         -> 
	   Tree.debug "%s > %s" (string_of_channel h) (string_of_channel k);
	   forwarder h k
       | #local_command -> failwith "forwarder received a illegal local command"
    )


(* code for localized ST ambients *)
let rec st_ambient h k s p =
  (* create sub-ambients,
   * add the process p to the local process, 
   * search for a potential match,
   * execute it or wait
   *)
  let _,s = store ~create s h 0 p in
  let s,m = get_match s in
    match m with
      | `None               -> st_ambient_rcv h k s
      | `Com (xl, paths, p) -> Stats.l_com(); st_ambient h k s (Subst.names xl paths p)
      | `In  (h', k', _)    ->  
	  Stats.l_in(); 
	  send_compl h' (`Go k');
	  send_compl k' (`OKin h); (* h is not used *)
	  st_ambient h k s zero
      | `Out (h', p)        -> 
	  Stats.l_out(); 
	  send_compl h' (`Go k);
	  st_ambient h k s p  
      | `Open (p, h')       -> 
	  Stats.l_open(); 
	  send_compl h' (`Migrate h); (* h is not used *)
	  wait_st_ambient h k s p  
      | `Req (r, p)         -> 
	  send_req k r;
	  wait_st_ambient h k s p  
      | `Print (w, p)       -> 
	  print_string w; print_newline();
	  st_ambient h k s p  
      | `Pause (r, w, p) when !r=0 -> wait_st_ambient h k s ~w p  
      | `Pause (r, w, p)    -> decr r; st_ambient h k s p  
      | `Sleep (t, p)       -> 
	  Thread.delay (float_of_int t);
	  st_ambient h k s p  
      | `DoOut _            -> failwith "DoOut in PAN mode."

(* reception of requests *)
and st_ambient_rcv h k s =
  `Receive
    (function
       | `R (req, j, _) -> Stats.request (j, List.length req); st_ambient h k (add_reqs s req) zero
       | `Decr
       | `Step _        -> st_ambient_rcv h k s
       | `C c           -> failwith "non-wait ambient received completion : `%a'" !!format_completion c
       | `Stop          -> stop_purge
       | `Collect c     -> ca c h s k; st_ambient_rcv h k s
       | `Debug         -> 
	   Tree.debug "%s _ %s : %a" (string_of_channel h) (string_of_channel k) !!format_state s;
	   st_ambient_rcv h k s
       | `Eval p        -> st_ambient h k s p
    )

and wait_st_ambient h k s ?w p =
  (* handle existing matches (ins),
   * execute it or wait *)
  match get_async_match s with
    | s, `In (h', k', _) -> 
	Stats.l_in(); 
	send_compl h' (`Go   k');
	send_compl k' (`OKin h);
	wait_st_ambient h k s ?w p
    | s, `None -> 
	wait_st_ambient_rcv h k s ?w p
    | s, `DoOut _ -> failwith "DoOut in PAN mode..."

(* reception of messages, in wait state (wait.p) *)
and wait_st_ambient_rcv h k s ?w p =
  `Receive
    (function
       | `R (req, j, _)  -> 
	   Stats.request (j, List.length req); 
	   wait_st_ambient h k (add_reqs s req) ?w p
       | `C (`Go k')     -> Stats.completion(); st_ambient h k' s p
       | `C (`OKin _)    -> Stats.completion(); st_ambient h k  s p
       | `C (`Migrate _) ->
	   Stats.completion(); 
	   Stats.forwarder (-1); 
	   send_compl k (`Register (true, p, s)); 
	   forwarder h k
       | `C (`Register (_, q, s')) ->
	   Stats.completion(); 
	   st_ambient h k (merge s h s') (par p q)
       | `Decr           -> wait_st_ambient_rcv h k s ?w p
       | `Stop           -> stop_purge
       | `Step "*" when w<>None   -> st_ambient h k s p
       | `Step w'  when w=Some w' -> st_ambient h k s p
       | `Step _                  -> wait_st_ambient ?w h k s p
       | `Collect c      -> ca c h s k; wait_st_ambient_rcv h k s ?w p
       | `Debug          -> 
	   Tree.debug "%s _ %s : %a@ wait.%a" (string_of_channel h) (string_of_channel k) !!format_state s !!Printer.format_process p;
	   wait_st_ambient_rcv h k s ?w p
       | `C _            -> failwith "wait ambient received a Gof"	 
       | #local_command  -> failwith "wait ambient received an illegal local command"
    )


(* creation of an ambient, named n, with father h (or itself by default) or host.root, 
 * with content p *)
and create ?host ?h ?(n=rootname) p = 
  let parent f k = match host with
    | None      -> f k (oget h k)
    | Some host -> 
	send_command ~host (`Register k);
	`Receive (function `C (`Go h) -> f k h | _ -> failwith "bad wait response") 
  in
    if immobile n then
      channel (parent (fun k h -> Immobile.ambient ~create k h (empty_state n) p)), 0
    else 
      channel (parent (fun k h -> st_ambient k h (empty_state n) p)), 0
