(*pp camlp4o *)

(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Parsing *)

open Ast

(* table of definitions *)
let defs = Hashtbl.create 5
let mkdef f vl p = Hashtbl.replace defs f (vl, p)
let deftype f = 
  try fst (Hashtbl.find defs f) 
  with Not_found -> raise (Stream.Error ("Unbound definition : `"^f^"'"))
let getdef f s = 
  let p = snd (Hashtbl.find defs f) in
  let n,r = List.fold_left 
	      (fun (n,r) -> function
		 | `R x -> n, x::r
		 | `N (x,y) -> (avar x,y)::n, r
	      ) ([],[]) s in
    List.fold_left (fun p (v,q) -> Subst.process v q p) (Subst.names' ~abs:true n p) r


(* we need to remember the type (name-path/recursion) of variables *)
let use t x =
  try
    if List.assoc x t <> `N then
      raise (Stream.Error ("Recursion variable `"^x^"' used as a name or a path"));
    avar x
  with Not_found -> x
let ruse t x =
  try match List.assoc x t with
    | `R v -> Var v
    | `N   -> Pref(Name (avar x), zero)
  with Not_found -> raise (Stream.Error ("Unbound variable : `"^x^"'"))
let rec ndecl t = function
  | [] -> t
  | w::q -> ndecl ((w,`N)::t) q
let rdecl t w x = (w,`R x)::t

let nilparser x = parser [< >] -> x

let capa t = parser
  | [< 'CAPA (Pause r); c=
	 (parser
	    | [< 'WORD w; c=
		   (let w = use t w in parser 
		      | [< 'INT i >] -> r:=i; Capa(Pause r, w)
		      | [< >]        ->       Capa(Pause r, w)
		   ) >] -> c
	    | [< 'INT i; c= 
		   (r:=i; parser 
		      | [< 'WORD w >] -> Capa(Pause r, use t w)
		      | [< >]         -> Capa(Pause r, "")
	      ) >] -> c
	    | [< >] -> Capa(Pause r, "")) >] -> c
  | [< 'CAPA c; 'WORD w ?? "Name expected after "^(string_of_cap c) >] -> Capa (c, use t w)
  | [< 'SLEEP; i=
	 (parser 
	    | [< 'INT i >] -> if i<1 then raise (Stream.Error "Sleep argument must be positive"); i 
	    | [< >] -> 1) >] -> Sleep i

(* WORD + capa *)
let expr t = parser
  | [< 'WORD w >] -> Name (use t w)
  | [< c = capa t >] -> c

(* expr* *)
let path t = 
  let rec aux = parser
    | [< 'DOT; e=expr t ?? "Capacity or variable expected after `.'"; q=aux >] -> e::q
    | [<  >] -> []
  in parser [< e=expr t; q=aux >] -> e::q | [< >] -> []

(* path* *)
let paths t = 
  let rec aux = parser [< 'COMMA; p=path t; q=aux >] -> p::q | [< >] -> []
  in parser [< p=path t; q=aux >] -> p::q | [< >] -> []

(* (COMMA name)* *)
let rec name_rest = parser
  | [< 'COMMA; 'WORD x ?? "Name expected after ','";
               q = name_rest >] -> x::q
  | [< >] -> []
let name_list = parser
  | [< 'WORD x; q=name_rest >] -> x::q
  | [< >] -> []

(* (COMMA name)* RPAR *)
let rec name_rest_rpar = parser
  | [< 'COMMA; 'WORD x ?? "Name expected after ','";
               q = name_rest_rpar ?? "`)' or ',' expected " >] -> x::q
  | [< 'RPAR >] -> []

(* a process: parallel composition of non-empty processes *)
let rec process t = parser
  | [< p = sprocess t; q = process_rest t >] -> p @ q
  | [< >] -> []
and process_rest t = parser
  | [< 'BAR; p = process t >] -> p
  | [< >] -> []

(* non-empty process *)
and sprocess t = parser
  | [< 'ZERO >] -> []
  | [< 'WORD w; p=wprocess t w >] -> p
  | [< 'LPAR; p=lprocess t ?? "Process expected after `('" >] -> p
  | [< 'LPIK; p=paths t; 'RPIK ?? "`>' expected" >] -> [Send p]
  | [< 'REC; 'WORD w ?? "Name expected after `rec'";
             'DOT    ?? "`.' expected after `rec "^w^"'";
	     x=nilparser (var w);
	     p=sprocess (rdecl t w x) >] -> [Rec(x, p)]
  | [< 'BANG; c=capa t  ?? "Capacity expected after `!'"; 
              p=cwprocess t c >] -> [Rec(bang_var,[Pref(c,Var bang_var::p)])]
  | [< c=capa t; p=cwprocess t c >] -> [Pref(c,p)]

(* the capacity c has been read *)
and cwprocess t c = parser
  | [< 'DOT; p=sprocess t ?? "Process expected after `.'" >] -> p
  | [< >] -> []

(* the name w has been read *)
and wprocess t w = parser
  | [< 'DOT; w=nilparser (use t w); p=sprocess t ?? "Process expected after `.'" >] -> [Pref(Name w, p)]
  | [< 'HOST h; p=whprocess t (use t w) h ?? "`:' or `[' expected after `n@host'" >] -> [p]
  | [< 'LPAR; s=args w t (deftype w) >] -> getdef w s
  | [< p=bprocess t w >] -> [Amb(use t w, p)]
  | [< >] -> [ruse t w]

(* waiting for the content of an ambient *)
and bprocess t w = parser
  | [< 'LSQR; p=process t; 'RSQR ?? "`]' expected" >] -> p

(* w@h has been read, waiting for the port (:i) or the content *)
and whprocess t w h = parser 
  | [< 'COLON; 'INT i ?? "Number expected after `n@host:'";
               p=bprocess t w ?? "`[' expected after `n@host:port'" >] -> Amb'((h, i), w, p)
  | [< p=bprocess t w >] -> Amb'((h, 1550), w, p)

(* a parenthesis has been opened *)
and lprocess t = parser
  | [< 'WORD w; p=lwprocess t w >] -> p
  | [< 'NU; 'WORD w ?? "Name expected after `nu"; 
            q=name_rest; 'RPAR   ?? "`)' expected"; 
	    p=sprocess t (*ndecl t (w::q)*) ?? "Process expected after `(nu ..)'" >] -> curry_nus p (w::q)
  | [< p=sprocess t; q=process_rest t; 'RPAR ?? "`)' expected" >] -> p @ q

(* a parenthesis has been opened, followed by a name w *)
and lwprocess t w = parser
  | [< wq=name_rest_rpar; p=pprocess (ndecl t (w::wq)) >] -> [Rcv(List.map avar (w::wq), p)]
  | [< p=wprocess t w; q=process_rest t; 'RPAR ?? "`)' expected" >] -> par p q

(* "weak" process: don't grab parallel composition without parenthesis *)
and pprocess t = parser
  | [< p=sprocess t >] -> p
  | [< >] -> []

(* definition arguments (depends on the definition type) *)
and args f t = 
  let arg = function
    | _,`R v -> (parser [< q=process t >] -> `R (v,q))
    | x,`N   -> (parser [< 'WORD n >] -> `N (x,[Name (use t n)]) | [< >] -> raise (Stream.Error ("Definition `"^f^"' expects a name there")))
  in
    function
      | [] -> (parser [< 'RPAR >] -> [] | [< >] -> raise (Stream.Error "`)' expected (definition without arguments)"))
      | [x] -> (parser [< s=arg x; 'RPAR ?? "`)' expected (mistyped use of definition `"^f^"')" >] -> [s])
      | x::q -> (parser [< s=arg x; 'COMMA ?? "`,' expected (mistyped use of definition `"^f^"')"; sq=args f t q >] -> s::sq)

(* definition parameters *)
let param_list = 
  let param = parser
    | [< 'QUOTE; 'WORD x >] -> x, `R (var x)
    | [< 'WORD n >]         -> n, `N
  in
  let rec rest = parser
    | [< 'COMMA; x=param ?? "parameter expected after `,'"; q=rest >] -> x::q
    | [< >] -> []
  in parser [< x=param; q=rest >] -> x::q | [< >] -> []

(* definition *)
let definition = parser
  | [< 'LET; 'WORD f ?? "Name expected after `let'";
             'LPAR   ?? "`(' expected after `let "^f^"'";
             env=param_list; 
	     'RPAR   ?? "`)' expected";
	     'EQUAL  ?? "`=' expected";
	     p=process env >] -> mkdef f env p; f
      
(* main entry: definitions, followed by a process *)
let rec main = parser
  | [< _=definition; p=main >] -> p
  | [< p=process [] >] -> p

(* cleaning *)
let rec clean e = parser
  | [< 'EOF  >] -> raise e
  | [< 'SEMI >] -> raise e
  | [< '_ ; x = clean e >] -> x


(* host:port *)
let host = parser
  | [< 'WORD h; p=(parser [< 'COLON; 'INT p ?? "portnumber expected" >] -> p | [< >] -> 1550) >] -> h,p

(* errors *)
let error l m = 
  let p = Lexing.lexeme_start_p l in
    ParseError (p.Lexing.pos_lnum, p.Lexing.pos_cnum-p.Lexing.pos_bol, m)


(* parsing from a closed lexbuf *)
let parse_lexbuf l =
  try 
    let f = Stream.from (fun _ -> Some (Lexer.token l)) in
      (parser [< p=main; 'EOF ?? "End of file expected" >] -> p) f
  with Stream.Error m -> raise (error l m)

(* parsing of a string *)
let from_string s = parse_lexbuf (Lexing.from_string s)

(* parsing of a channel *)
let from_channel i = parse_lexbuf (Lexing.from_channel i)

(* parsing of a file *)
let from_file   s = 
  let i = open_in s in
    try let p = from_channel i in close_in i; p
    with e -> close_in i; raise e


(* interactive parsing *)
let interactive i =
  let b = Lexing.from_channel i in
  let f = Stream.from (fun _ -> Some (Lexer.token b)) in
  let e = parser 
    | [< 'SEMI >] -> () 
    | [< _ = clean (error b ("Unexpected token `"^Lexing.lexeme b^"'")) >] -> ()
  in
  let bool = parser 
    | [< 'WORD "on"   >] -> true
    | [< 'WORD "off"  >] -> false
    | [< >]              -> true
  in
  let iparse = parser 
    | [< 'DIESE "debug" >] -> `Cmd `Debug
    | [< 'DIESE "tree" >] -> `Cmd `Tree
    | [< 'DIESE "add"; p=process [] >] -> `Add p
    | [< 'DIESE "step";  w=
	   (parser 
	      | [< 'STAR   >] -> "*"
	      | [< 'WORD w >] -> w
	      | [< >]         -> ""
	   ) >] -> `Cmd (`Step w)
    | [< 'DIESE "safenames"; b=bool >] -> `SafeNames b
    | [< 'DIESE "packrequests"; b=bool >] -> `PackReq b
    | [< 'DIESE "relocation"; b=
	   (parser 
	      | [< 'WORD "sync"  >] -> `Sync
	      | [< 'WORD "async" >] -> `Async
	      | [< 'WORD "mixed" >] -> `Mixed
	   ) ?? "`sync', `async' or `opt' expected." >] -> `RelocMode b
    | [< 'DIESE "stats" >] -> `Stats
    | [< 'DIESE "print"; p=process [] >] -> `Print p
    | [< 'DIESE "addto"; h=host ?? "host:port expected after #addto"; 
         p=process [] ?? "single ambient expected after #addto" >] -> 
	(match p with 
	   | [Amb(n,p)] -> `AddTo (h,n,p) 
	   | _ -> raise (Stream.Error ("Single ambient expected with #addto")))
    | [< 'DIESE "check"; p=process [] >] -> `Check p
    | [< 'DIESE "help" >] -> `Help
    | [< 'DIESE d >] -> raise (Stream.Error ("Unknown directive : `"^d^"'"))
    | [< f=definition >] -> `Def f
    | [< p=process [] >] -> `Eval p
  in      
  let parse = parser
    | [< 'EOF >] -> None
    | [< i = iparse; _=e >] -> Some i
  in
  let rec get x = 
    let l0 =  (Lexing.lexeme_start_p b).Lexing.pos_lnum in 
      try 
	try parse f with Stream.Error m -> clean (error b m) f
      with ParseError (l,c,m) -> 
 	if l=l0 then Printf.printf "Parse error at character %d : %s\n%!" c m
 	else Printf.printf "Parse error at line %d, character %d : %s\n%!" (l-l0) c m;
	get x
  in
    Stream.from get


(* Prints help (list of directives) *)
let print_help() = Printf.printf
"                               GcPan Help

 (further information is available at http://perso.ens-lyon.fr/gcpan/)

- sentences end with a `;;'

- a sentence is either:
  * a well-typed Safe Ambient process: the environment is cleared and
    P gets evaluated.
  * a macro definition `let f(a,'P) = ...'
    the variable `a' ranges over names while `P' ranges over processes,
    due to the prepending quote
  * a directive, `#directive args'

- Here are the directives:
 #add P           : add P to the current environment
 #addto host a[P] : register P at the given host, but execute it locally
 #step / #step p  : unblock all local threads waiting with pause.P / pause p.P
 #print P         : print P (with all macros expanded)
 #tree            : print the current tree,
                  forwarders are represented with `<[P]', 
                  some trees may appear disconnected when they are
                  waiting for a completion message, and that their old
                  parent location has been garbage collected
 #stats           : print the current statistics
 #check P         : type-check P (experimental), 0 means not typable
 #help            : print this help
 #debug           : print the full state of each location
 #safenames {on,off}
                  : when distributing the code over several instances of
                  gcpan one has to be careful to fresh names (nu n).
                  By default a fresh name is created using the name of the
                  site, and a UID defined by this site.
                  This leads to long names, and is not required when there
                  is only one instance of gcpan. This directive allows you
                  to disable safe names when you play with only one instance.
 #packrequests {on,off}
                  : {en,dis}able the optimisation that consists in packing 
                  requests messages when a forwarder is blocked
 #relocation {sync,async,opt}
                  :  select the strategy for forwarders relocation
                     * sync means that the relocation message goes sequentially
                       along the forwarders chain
                     * async means that the relocation message is broadcasted
                       to the chain
                     * opt should be the optimal merging of these two
                       strategies: the relocation message is broadcasted only
                       once by physical location, and then routed sequentially
                       inside each physical location.
%!"


(* Prints version and licence message *)
let print_version() = Printf.printf
"GcPan 13/10/2004, Copyright (C) 2004 Damien Pous
       http://perso.ens-lyon.fr/gcpan/

GcPan is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

GcPan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GcPan; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%!"
