(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Tools to collect/print a frozen view of the local net *)

open Ast

class type collector =
object
  method ambient: string -> string -> string -> unit
  method forwarder: string -> string -> unit
  method dead: unit
end

class type ops =
object
  method debug: unit
  method tree : process
  method ftree: process
  method print_tree : unit
  method print_ftree: unit
end


(* lock *)
class l =
  let m = Mutex.create () in
  let f = Condition.create () in
  let n = ref 0 in
object
  method lock = 
    Mutex.lock m;
    if !n > 0 then Condition.wait f m (* another collect is executing *)

  method locked (v: unit -> unit) = 
    Mutex.lock m;
    v();
    decr n;
    if !n > 0 then Condition.wait f m
    else Condition.broadcast f;
    Mutex.unlock m;    

  method wait i = 
    n := i;
    if i > 0 then Condition.wait f m (* wait i responses *)    

  method unlock = 
    Mutex.unlock m
end
let l = new l


(* debugger *)
let debug f = Format.kprintf (fun s -> l#locked (fun()->print_endline s)) f


(* tree collector *)
class c =
  let children = Hashtbl.create 6 in
  let agents   = Hashtbl.create 6 in
  let get_children k = try Hashtbl.find children k with Not_found->[] in
  let fold aux k = List.fold_left (fun q h -> par (aux h) q) zero (get_children k)
  in
object(s)
  val mutable root=""
  method private add k h x = 
    if k<>h then (
      Hashtbl.replace children k (h::get_children k); 
      Hashtbl.add agents h x; 
    ) else root <- h

  method ambient h n k = l#locked (fun()->s#add k h (`Amb n))
  method forwarder h k = l#locked (fun()->s#add k h (`Fwd))
  method dead          = l#locked (fun()->())

  method private roots = 
    Hashtbl.fold (fun k _ q -> if Hashtbl.mem agents k then q else k::q) children [] 

  method private return aux =
    let p = List.fold_left (fun q h -> par (aux h) q) zero s#roots in
      Hashtbl.clear children;
      Hashtbl.clear agents;
      root <- "";
      p

  method tree  = 
    let rec aux k = 
	try match Hashtbl.find agents k 
	with
	  | `Amb n -> [Amb(n, fold aux k)]
	  | `Fwd   -> fold aux k 
	with Not_found -> [Amb((if k=root then rootname else "?"), fold aux k)]
    in
      s#return aux

  method ftree = 
    let rec aux k = 
	try match Hashtbl.find agents k 
	with
	  | `Amb n -> [Amb(n, fold aux k)]
	  | `Fwd   -> [Amb("<", fold aux k)]
(* 	  | `Fwd   -> [match fold aux k with *)
(* 			 | [] -> Pref(Name "", [Pref(Name "", [])]) *)
(* 			 | l  -> Pref(Name "", l)] *)
	with Not_found -> [Amb((if k=root then rootname else "?"), fold aux k)]
    in s#return aux

  method print_tree  = List.iter Printer.print_process_t s#tree
  method print_ftree = List.iter Printer.print_process_t s#ftree
end
let c = new c


(* operations *)
class o broadcast: ops =
object
  method debug = 
    l#lock; l#wait (broadcast `Debug); l#unlock
  method tree  = 
    l#lock; l#wait (broadcast (`Collect (c:>collector)));
    let p = c#tree in
      l#unlock;
      p
  method ftree = 
    l#lock; l#wait (broadcast (`Collect (c:>collector)));
    let p = c#ftree in
      l#unlock;
      p
  method print_tree  = 
    l#lock; l#wait (broadcast (`Collect (c:>collector)));
    c#print_tree;
    l#unlock
  method print_ftree = 
    l#lock; l#wait (broadcast (`Collect (c:>collector)));
    c#print_ftree;
    l#unlock
end
let collect = new o
