(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of GcPan.
 * 
 * GcPan is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GcPan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GcPan; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

(* Type inference *)


open Ast

(* Type variables *)
module V = Names.Make(struct let base="v" end) (* names, abstractions *)
module P = Names.Make(struct let base="p" end) (* path variables *)
module R = Names.Make(struct let base="x" end) (* recursion variables *)
module I = Names.Make(struct let base="'" end) (* thread variables *)


(* module Unif: sig *)
(*   type ('a,'b) t *)
(*   val create: unit -> ('a,'b) t *)
(* (\*   val format: (Format.fomatter -> 'a -> unit) -> (Format.fomatter -> 'b -> unit) -> Format.fomatter -> ('a,'b) t -> unit *\) *)
(*   val insert:  ('a,'b) t -> 'a -> 'b -> t *)
(*   val define:  ('a,'b) t -> 'a -> 'b -> t *)
(*   val find  :  ('a,'b) t -> 'a -> 'b *)
(*   val undef :  ('a,'b) t -> 'a -> t *)
(* end = struct *)
(*   type ('a,'b) t = {  *)
(*     binds: 'a*'b list; *)
(*     defs : 'a*'b list; *)
(*   } *)
(* end *)


let vfresh() = `V (V.fresh())
let pfresh() = `V (P.fresh())
let rfresh() = `V (R.fresh())
let ifresh() = `V (I.fresh())

(* Thread: true/false/variable *)
type i = 
    [ `T | `F 
    | `V of I.t ]
    
(* IM / ST *)
type r = 
    [ `IM of p 
    | `ST of i * i * p
    | `V  of R.t ]
      
(* Type of messages *)
and v = 
    [ `A of r 
    | `C of r 
    | `V of V.t ]
and p = 
    [ `L of v list
    | `V of P.t]


(* constraints *)
type c = (V.t * v) list * (P.t * p) list * (R.t * r) list * (I.t * i) list


(* insatisfiable constraint *)
exception False


(* base generic unification: variables only *)
let unify insert s (i,j) = match i,j with
  | `V v, `V v'       -> if v>v' then insert (v,j) else if v=v' then s else insert (v',i)
  | `V v, x | x, `V v -> insert (v,x)
  | _                 -> raise False

(* utility for the following unification functions:
 * when a substitution is found, we insert it into the accumulator,
 * if the variable was already substituted, we unify recursively 
 * the two instances *)
let insert unify build l = 
  fun (v,x as vx) ->
    let rec aux acc = function
      | [] -> build (List.rev (vx::acc))
      | ((w,_)::q) as wq when v<w -> build (List.rev_append acc (vx::wq))
      | ((w,y)::q)       when v=w -> unify x y
      | wy::q -> aux (wy::acc) q
    in aux [] l

(* unification of two thread types, return the accumulator (s) 
 * augmented with needed substitutions *)
let rec iunify (vs,ps,rs,is as s : c) i j: c = 
  match i,j with
    | `T, `T 
    | `F, `F -> s
    | ij     -> unify (insert (iunify s) (fun is -> vs,ps,rs,is) is) s ij
	  
(* unification of two message types, return the accumulators (ws,rs,is)) 
 * augmented with needed substitutions *)
let rec vunify (vs,ps,rs,is as s : c) x y = 
  match x,y with
    | `A x, `A y 
    | `C x, `C y -> runify s x y
    | xy         -> unify (insert (vunify s) (fun vs -> vs,ps,rs,is) vs) s xy

and punify (vs,ps,rs,is as s : c) x y =
  match x,y with
    | `L x, `L y -> (try List.fold_left2 vunify s x y with Invalid_argument _ -> raise False)
    | xy         -> unify (insert (punify s) (fun ps -> vs,ps,rs,is) ps) s xy
    
(* unification of process types, return the accumulators (ws,rs,is)) 
 * augmented with needed substitutions *)
and runify (vs,ps,rs,is as s : c) x y = 
  match x,y with
    | `IM x, `IM y           -> punify s x y
    | `ST(i,k,x), `ST(j,l,y) -> punify (iunify (iunify s i j) k l) x y
    | xy                     -> unify (insert (runify s) (fun rs -> vs,ps,rs,is) rs) s xy


(* filtrage d'une liste de types *)
let filter  f = List.fold_left (fun acc t -> try f t :: acc with False -> acc) []
let ffilter f = List.fold_left (fun acc t -> try List.rev_append (f t) acc with False -> acc) []

(* safe evaluation of a "lazy" list *)
let force l = List.fold_left (fun acc f -> try f() :: acc with False -> acc) [] l


(* type inference *)
let infer (ne,ve as e) =

  (* "type" of a name type variable *)
  let type_of (vs,_,_,_) n =
    try 
      let v = Hashtbl.find ne n in
      let rec aux = function
	| (w,_)::q when w<v -> aux q
	| (w,t)::q when w=v -> t
	| _ -> `V v
      in aux vs
    with 
	Not_found -> let v = V.fresh() in 
	  Hashtbl.add ne n v;
	  `V v
  in


  (* inference for a parallel composition *)
  let rec infer s = 
    let rec aux = function
      | `IM _ as t -> 
	  filter (function
		    | `ST _,_ -> raise False
		    | t',s -> t, runify s t t'
		 )
      | `ST(`T,j,wl) as t -> 
	  filter (function
		    | `IM _,_ -> raise False
		    | t',s -> t, runify s t' (`ST(`F,j,wl))
		 )
      | `ST(`F,j,wl) as t -> 
	  filter (function
		    | `IM _,_ -> raise False
		    | t',s -> let i=ifresh() in `ST(i,j,wl), runify s t' (`ST(i,j,wl))
		 )
      | `ST(i,j,wl) as t -> 
	  ffilter (function
		     | `IM _,_ -> raise False
		     | `ST(`F,j',wl'),s -> [t, punify (iunify s j j') wl wl']
		     | `ST(`T,j',wl') as t',s -> [t', runify s t (`ST(`F,j',wl'))]
		     | t',s -> 
			 (try let si'F = runify s t' (`ST(`F,j,wl)) in
			    force [
			      (fun()->`ST(`T,j,wl), iunify (runify s t' (`ST(`T,j,wl))) i `F);
			      (fun()->`ST(`T,j,wl), iunify  si'F i `T);
			      (fun()->`ST(`F,j,wl), iunify  si'F i `F);
			    ]
			  with False -> [`ST(`T,j,wl), iunify (runify s t' (`ST(`T,j,wl))) i `F])
		  )
      | t -> 
	  ffilter (function
		     | `IM _ as t',s -> [t', runify s t t']
		     | `ST(`F,j',wl'),s -> [t, runify s t (`ST(ifresh(),j',wl'))]
		     | `ST(`T,j',wl') as t',s -> [t', runify s t (`ST(`F,j',wl'))]
		     | `ST(`V _ as i',j,wl) as t', s -> 
			 (try let siF = runify s t (`ST(`F,j,wl)) in
			    force [
			      (fun()->`ST(`T,j,wl), iunify (runify s t (`ST(`T,j,wl))) i' `F);
			      (fun()->`ST(`T,j,wl), iunify  siF i' `T);
			      (fun()->`ST(`F,j,wl), iunify  siF i' `F);
			    ]
			  with False -> [`ST(`T,j,wl), iunify (runify s t (`ST(`T,j,wl))) i' `F])
		     | t', s -> 
			 let j,wl = ifresh(),pfresh() in
			   try 
			     let siF = runify s t (`ST(`F,j,wl)) in
			       force [
				 (fun()->`IM wl, runify (runify s t (`IM wl)) t' (`IM wl));
				 (fun()->`ST(`T,j,wl), runify (runify s t (`ST(`T,j,wl))) t' (`ST(`F,j,wl)));
				 (fun()->`ST(`T,j,wl), runify  siF t' (`ST(`T,j,wl)));
				 (fun()->`ST(`F,j,wl), runify  siF t' (`ST(`F,j,wl)));
			       ]
			   with False -> 
			     force [
			       (fun()->`IM wl, runify (runify s t (`IM wl)) t' (`IM wl));
			       (fun()->`ST(`T,j,wl), runify (runify s t (`ST(`T,j,wl))) t' (`ST(`F,j,wl)));
			     ]
		  )
    in function
      | []   -> [rfresh(), s]
      | p::q -> ffilter (fun (t,s) -> aux t (infer s q)) (infer_t s p) 
	  (* TOTHINK: a fait bcp d'appels  (infer s q) 
	   * (avec un s qui est srement peu diffrent  chaque fois...)
	   *)

  (* inference for a "single" process *)
  and infer_t s = function   
    | Amb (n,p) | Amb'(_,n,p) -> 
	(match type_of s n with
	   | `C _ -> raise False
	   | `A (`IM _ as t) -> 
	       filter (function
			 | `ST _, _ -> raise False
			 | t', s    -> rfresh(), runify s t t'
		      ) (infer s p)
	   | `A (`ST(_,j,wl)) ->
	       filter (function
			 | `IM _, _         -> raise False
			 | `ST(_,j',wl'), s -> rfresh(), punify (iunify s j j') wl wl'
			 | t', s            -> rfresh(), runify s t' (`ST(ifresh(),j,wl))
		      ) (infer s p)
	   | t -> 
	       ffilter (function
			  | `IM _ as t', s -> [rfresh(), vunify s t (`A t')]
			  | `ST(_,j,wl), s -> [rfresh(), vunify s t (`A (`ST(ifresh(),j,wl)))]
			  | t', s          -> 
			      let j,m=ifresh(),pfresh() in
				(* TRICHE: il ne faut garder que le premier cas !!! *)
				if is_abstract n then 
				  force [
				    (fun()-> rfresh(), runify (vunify s t (`A (`IM m))) t' (`IM m));
				    (fun()-> rfresh(), runify (vunify s t (`A (`ST(ifresh(),j,m)))) t' (`ST(ifresh(),j,m)));
				  ]
				else if immobile n then [rfresh(), runify (vunify s t (`A (`IM m))) t' (`IM m)]
				else [rfresh(), runify (vunify s t (`A (`ST(ifresh(),j,m)))) t' (`ST(ifresh(),j,m))]
		       ) (infer s p)
	)

    | Pref(e,p) -> 
	ffilter 
	  (function
	     | `IM _ as t, s -> 
		 filter (function
			   | `ST _, _ -> raise False
			   | t', s    -> t, runify s t t'
			) (infer s p)
	     | `ST(_,j,wl) as t, s -> 
		 filter (function
			   | `IM _, _ -> raise False
			   | t', s    -> `ST(`T,j,wl), runify s t t'
			) (infer s p)
	     | t, s -> 
		 filter (function
			   | `ST(_,j,wl) as t', s -> `ST(`T,j,wl), runify s t t'
			   | t', s -> t', runify s t t'
			) (infer s p)
	  ) (infer_expr s e)
	    
    | Send el -> 
	List.fold_left  
	    (fun acc (wl,s) -> let wl = `L wl in
	       (`IM wl, s) :: (`ST(`T,`T,wl), s) :: (`ST(ifresh(),`F,wl), s) :: acc) 
	  [] (infer_paths s el)

    | Rcv (xl,p) -> 
	let vl = `L (List.map (fun x -> let v=V.fresh() in Hashtbl.add ne x v; `V v) xl) in
	let t  = infer s p in
	let _  = List.iter (Hashtbl.remove ne) xl in
	  ffilter (function
		     | `IM wl, s      -> [`IM wl, punify s vl wl]
		     | `ST(i,j,wl), s -> [`ST(ifresh(),j,wl), punify (iunify s i j) vl wl]
		     | `V _ as t, s   -> force [
			 (fun()-> `IM vl, runify s t (`IM vl));
			 (fun()-> let i = ifresh() in
				 `ST (ifresh(),i,vl), runify s t (`ST (i,i,vl)))
		       ]
		  ) t
    
    | Nu  (n,p) ->
	let v = V.fresh() in
	let _ = Hashtbl.add ne n v in
	let t = infer s p in
	let _ = Hashtbl.remove ne n in
	  t
	
    | Rec (x,p) -> 
	let r = rfresh() in
	let _ = Hashtbl.add ve x r in
	let t = infer s p in
	let _ = Hashtbl.remove ve x in
	  filter (fun (t,s) -> t, runify s r t) t

    | Var  x      -> 
	[Hashtbl.find ve x, s]


  (* inference for a set of paths *)
  and infer_paths s = function
    | [] -> [[], s]
    | p::q -> ffilter (fun (tq,s) -> List.map (fun (t,s) -> t::tq, s) (infer_path s p)) (infer_paths s q)

  and infer_path s = function
    | [Name n] -> let r=rfresh() in force [
	(fun()-> `C r, vunify s (type_of s n) (`C r));
	(fun()-> `A r, vunify s (type_of s n) (`A r));
      ]
    | l        -> infer_real_path s l
  and infer_real_path s = function
    | []   -> [`C (rfresh()), s]
    | p::q -> ffilter (function
			 | `A _,_ -> raise False
			 | `C (`IM wl) as t, s -> 
			     filter (function 
				       | `ST _,_ -> raise False
				       | t', s -> t, runify s t' (`IM wl)
				    ) (infer_expr s p) 
			 | `C (`ST(i,j,wl)) as t, s -> 
			     filter (function 
				       | `IM _,_ -> raise False
				       | t', s -> t, runify s t' (`ST(`T,j,wl))
				    ) (infer_expr s p) 
			 | t, s -> 
			     ffilter (function 
					| `IM _ as t',s -> [`C t', vunify s t (`C t')]
					| `ST(i,_,_) as t',s -> [`C t', vunify (iunify s i `T) t (`C t')]
					| t',s -> force [
					    (fun()-> let t''=`C (`IM (pfresh())) in t'', vunify (vunify s t t'') (`C t') t'');
					    (fun()-> 
					       let i,j,wl=ifresh(),ifresh(),pfresh() in
					       let t''=`C (`ST(i,j,wl)) in 
						 t'', runify (vunify s t t'') t' (`ST(`T,j,wl)));
					  ] 
				     ) (infer_expr s p) 
		      ) (infer_real_path s q)


  (* inference for an expression *)
  and infer_expr s = function
    | Name  n       -> let r=rfresh() in [r, vunify s (type_of s n) (`C r)] 
    | Capa(In,   n) 
    | Capa(Out,  n) -> [`ST(`T,ifresh(),pfresh()), vunify s (type_of s n) (`A (rfresh()))]
    | Capa(Open_,n) -> let t=`ST(ifresh(),ifresh(),pfresh()) in [t, vunify s (type_of s n) (`A t)]
    | Capa(In_,  n) 
    | Capa(Out_, n) -> (* TRICHE: ne garder que le premier cas *)
	if is_abstract n then
	  force [
	    (fun()-> `IM (pfresh()),            vunify s (type_of s n) (`A (`IM (pfresh()))));
	    (fun()-> `ST(`T,ifresh(),pfresh()), vunify s (type_of s n) (`A (`ST(ifresh(),ifresh(),pfresh()))));
	  ]
	else if immobile n then [`IM (pfresh()), vunify s (type_of s n) (`A (`IM (pfresh())))]
	else [`ST(`T,ifresh(),pfresh()), vunify s (type_of s n) (`A (`ST(ifresh(),ifresh(),pfresh())))]

    | Capa(Open, n) -> 
	let j,wl = ifresh(),pfresh() in 
	let tf, tt = `ST(`F,j,wl), `ST(`T,j,wl) in
	force [
	  (fun()-> `IM wl, vunify s (type_of s n) (`A (`ST(`F,`F,wl))));
	  (fun()->  tf   , vunify s (type_of s n) (`A tt));
	  (fun()->  tt   , vunify s (type_of s n) (`A tf));
	]
    | _             -> [rfresh(), s]
	
  in infer 

let format_i f = function
  | `T   -> Format.pp_print_char f '^'  
  | `F   -> Format.pp_print_char f ''
  | `V i -> I.format f i

let rec format_r f = function
  | `IM m      -> Format.fprintf f "IM[%a]" format_p m
  | `ST(i,j,m) -> Format.fprintf f "ST%a[@[%a%a@]]" format_i i format_i j format_p m
  | `V r       -> R.format f r

and format_v f = function
  | `A r -> Format.fprintf f "Amb%a" format_r r  
  | `C r -> Format.fprintf f "Cap%a" format_r r
  | `V v -> V.format f v

and format_p f = function
  | `L l -> Printer.format_list format_v f l
  | `V p -> P.format f p
